"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import time
import nanosurf.lib.devices.i2c.bus_master as i2c 

class Chip_24LC32A(i2c.I2CChip):
    """ EEPROM chip"""
    def __init__(self, bus_addr: int, **kwargs):
        super().__init__(bus_addr, offset_mode=i2c.I2COffsetMode.U16Bit_MSBFiRST, **kwargs)

    def memory_read_bytes(self, base_addr: int, len: int) -> list[int]:
        return self.read_bytes_with_offset(base_addr, len)

    def memory_write_bytes(self, base_addr: int, data: list[int]):
        page_size = 32
        page_base_addr = int(base_addr / page_size)

        # make sure the data array is block aligned, in base addr and size
        data_to_write = []

        # if not fill it up at the beginning
        page_base_start_offset = base_addr % page_size
        if page_base_start_offset > 0:
            first_page_data = self.read_bytes_with_offset(page_base_addr, page_base_start_offset)
            data_to_write = first_page_data.append(data)

        # fill up at the end
        missing_bytes = len(data) % page_size
        if missing_bytes > 0:
            missing_page_data = self.read_bytes_with_offset(page_base_addr+len(data), missing_bytes)
            data_to_write.append(missing_page_data)

        # check if we did correctly prepare the data
        assert (page_base_addr % page_size) == 0, "Error: base addr is not page aligned"
        assert (len(data) % page_size) == 0, "Error: data array size is not multiple of page size"
        assert data_to_write[page_base_addr-base_addr:page_base_addr-base_addr+len(data)] is data, "Error: data alignment failed"

        # writting data page by page
        bytes_to_send = len(data)
        pages_to_write = int(bytes_to_send / page_size)
        for current_page in range(pages_to_write):
            current_page_addr = page_base_addr + current_page*page_size
            current_data_start = current_page*page_size
            current_data_stop = (current_page+1)*page_size
            # self.write_bytes_with_offset(current_page_addr, data[current_data_start:current_data_stop])
            time.sleep(0.5) # is needed to let the EEPROM finish internal write cycle




"""Parser for LogicalUnit_InterfaceShared.h
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import re

class ParseShared():

    def __init__(self, lu_shared_file_path, verbose=False):
        self._lu_file = open(lu_shared_file_path, 'r')
        self._verbose = verbose
        self.lu_type_definitions = dict()
        self._parse_lu_shared_file()
        self._lu_file.close()

    def __del__(self):
        self._lu_file.close()

    def _parse_enum(self):
        state_wait_for_enum = 1
        state_wait_for_open_brace = 2
        state_read_ID = 3
        enum_name = ""
        enum_dict = {}
        state = state_wait_for_enum
        for lu_line in self._lu_file:
            # ignore comments and emty spaces
            lu_line = lu_line.split("//")[0].strip()

            if state == state_wait_for_enum:
                curline = lu_line.split(" ")
                if curline[0] == "enum":
                    enum_name = curline[1]
                    # print("enum = " + enum)
                    state = state_wait_for_open_brace
                    continue

            if state == state_wait_for_open_brace:
                curline = lu_line.split(" ")
                if curline[0] == "{":
                    index = -1
                    state = state_read_ID
                    continue

            if state == state_read_ID:
                curline = lu_line.split(",")
                if curline[0] == "":
                    # print("Skip empty line")
                    pass
                elif "}" in curline[0]:
                    # print("End of block found")
                    break
                else:
                    curvalue = curline[0].strip().split("=")
                    # print(curvalue)
                    name = curvalue[0].strip()
                    if name in ('BEGIN', 'END', 'NumInstances'):
                        continue

                    if len(curvalue) > 1:
                        index = int(curvalue[1].strip().split("/")[0], 0)
                        # print("index = " + str(index))
                    else:
                        index = index + 1
                        # print("AutoIndex = " + str(index))
                    enum_dict.update({name: index})
                    continue
        return enum_name, enum_dict

    def _clear_dictionaries(self):
        self._instances = dict()
        self._attribute_dict = dict()
        self._attribute_enum_dict = dict()
        self._trigger_dict = dict()
        self._busy_dict = dict()

    def _add_dictionaries(self, lu_type):
        if len(self._instances) > 0:
            self.lu_type_definitions[lu_type] = {
                'id': self.lu_type_dict[lu_type],
                'instances': self._instances,
                'attributes': self._attribute_dict,
                'attribute_enumerations': self._attribute_enum_dict,
                'triggers': self._trigger_dict,
                'busy': self._busy_dict
            }

    def _parse_lu_shared_file(self):
        def print_warning(message):
            if (self._verbose):
                print("Warning: Enumeration \"" + name + "\": "
                      + message
                )

        (name, self.lu_type_dict) = self._parse_enum()
        if name != 'LUTypeID':
            print_warning("First enumeration must be \"LUTypeID\"!")
            raise ValueError

        last_lu_type = ''
        current_lu_type = ''
        self._clear_dictionaries()

        ParseState = enum.Enum(
            'ParseState',
            ['instances', 'attributes', 'triggers', 'attribute_enum'])
        state = ParseState.instances
        while True:
            (name, dictionary) = self._parse_enum()
            if name == "":
                self._add_dictionaries(last_lu_type)
                break

            lu_enum_type_match = re.match(r"LU([a-zA-Z]+)_(\w+)", name)
            if lu_enum_type_match is None:
                print_warning("Skip, Enumeration name not parsable as LU enumeration type.")
                continue

            current_lu_type = lu_enum_type_match.group(1)
            if current_lu_type not in self.lu_type_dict:
                print_warning("Skip, does not match an LU type from LUTypeID.")
                continue

            if current_lu_type != last_lu_type:
                self._add_dictionaries(last_lu_type)
                self._clear_dictionaries()
                last_lu_type = current_lu_type
                state = ParseState.instances

            enum_type = lu_enum_type_match.group(2)

            if state == ParseState.instances:
                if enum_type != "InstID":
                    print_warning("Instance ID enumeration expected !")
                    raise ValueError

                self._instances = dictionary
                state = ParseState.attributes
                continue

            if state == ParseState.attributes:
                if enum_type != "AttrID":
                    print_warning(f"Attribute enumeration for {last_lu_type} expected."
                    )
                    raise ValueError

                self._attribute_dict = dictionary
                state = ParseState.triggers
                continue

            if state == ParseState.triggers:
                if enum_type != "TriggerID":
                    print_warning(f"Trigger enumeration for {last_lu_type} expected.")
                    raise ValueError

                self._trigger_dict = dictionary
                state = ParseState.attribute_enum
                continue

            if state == ParseState.attribute_enum:
                if enum_type == "BusyID":
                    # busy is an optional, pre-defined type
                    self._busy_dict = dictionary
                else:
                    # add attribute value enumeraton to ???
                    self._attribute_enum_dict[enum_type] = dictionary
                    state = ParseState.attribute_enum
                continue


if __name__ == "__main__":
    print("\nParse test")
    my_test = ParseShared(r"../../test/LogicalUnit_InterfaceShared.h", True)



# ///
# /// Line Edit for advanced number formatting
# ///
# /// Copyright (C) Nanosurf AG - All Rights Reserved (2021)
# /// Unauthorized copying of this file, via any medium is strictly prohibited
# /// https://www.nanosurf.com
# ///


from PySide2 import QtGui, QtWidgets
from PySide2.QtCore import Qt, Signal

class _LineEditQt(QtWidgets.QLineEdit):

    value_changed_event = Signal(float)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._block_value_change_signal = False
        self.textEdited.connect(self._value_changed())

    # public interface ------------------------------------------

    def set_value(self, new_value: str, notify: bool = True):
        do_notify = notify and (new_value != self.text())
        self.setText(new_value)
        if do_notify:
            self._value_changed()

    def value(self) -> str:
        return self.text()  

    # internal functions --------------------------------------------    

    # def focusInEvent(self, event: QtGui.QFocusEvent) -> None:
    #     """ overwrite handler for widget focus events"""
    #     super().focusInEvent(event)

    def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
        """  overwrite handler for widget focus events"""
        super().focusOutEvent(event)
        self._value_changed()

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        """ overwrite handler for key events"""
        if self.isReadOnly():
            return
        shift = (event.modifiers() == Qt.ShiftModifier)
        ctrl = (event.modifiers() == Qt.ControlModifier)
        key = event.key()
        if (key == Qt.Key_Return) or (key == Qt.Key_Enter):
            super().keyPressEvent(event)
            self._value_changed()
        else:
            super().keyPressEvent(event)

    def _value_changed(self):
        if not(self._block_value_change_signal):
            self.value_changed_event.emit(self.value())    


class NSFEdit(QtWidgets.QWidget):
    """ Custom Qt Widget to show text in a edit with a descriptive label. """
    value_changed_event = Signal(str)

    def __init__(self, label_str = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_widgets(label_str)
        self._edit.value_changed_event.connect(self._on_value_changed)
 
    def set_value(self, text: str, notify: bool = True):
        self._edit.set_value(text)
        if notify:
            self._on_value_changed()        
       
    def set_label(self, label: str):
       self._label.setText(label) 

    def set_read_only(self, set_read_only: bool):
        self._edit.setReadOnly(set_read_only)
        self._edit.setDisabled(set_read_only)
   
    def read_only(self) -> bool:
        return self._edit.isReadOnly()

    def label(self) -> str:
        return self._label.text()

    def value(self) -> str:
        return self._edit.value()  


    # internal ----------------------------------------------------
    
    def _setup_widgets(self, label_str: str):
        layout = QtWidgets.QVBoxLayout()
        self._label = QtWidgets.QLabel()
        self._label.setText(label_str)
        self._label.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._label, alignment=Qt.AlignBottom)
        self._edit = _LineEditQt()
        self._edit.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._edit)#, alignment=Qt.AlignTop)
        self.setLayout(layout)

    def _on_value_changed(self):
        self.value_changed_event.emit(self.value())




""" Some helpful mathematical functions for analyzing data in SciChannels and SciStreams
Copyright Nanosurf AG 2021
License - MIT
"""
import enum
import math
import numpy as np
import scipy.signal
import statistics as stat 
import nanosurf.lib.datatypes.sci_channel as ch
import nanosurf.lib.datatypes.sci_stream as ss

g_use_matplotlib_format_for_unit_str = False

def get_spectral_density_unit() -> str:
    return r"/$\sqrt{Hz}$" if g_use_matplotlib_format_for_unit_str else r"\sqrt(Hz)"

def find_peaks(stream: ss.SciStream, channel_index: int = 0, **kwargs) -> ss.SciStream:
    """ Find all peaks in a stream's channel. e.g amplitude peaks in a frq-spectrum.
    It is based on scipy.signal.find_peaks() for detailed documentation  
    
    Returns
    -------
    SciStream : 
        x-data contains the x positions of the peaks
        data stream contains the y positions of the peaks
    """
    data_channel = stream.get_channel(channel_index)
    peak_indexes, _ = scipy.signal.find_peaks(data_channel.value, **kwargs)    
    max_x_array = np.empty(len(peak_indexes))
    max_y_array = np.empty_like(max_x_array)
    for idx, peak_index in enumerate(peak_indexes):
        max_x_array[idx] = stream.x.value[peak_index]
        max_y_array[idx] = data_channel.value[peak_index]
    
    # create resulting stream
    result = ss.SciStream(channels=1)
    result.x = ch.SciChannel(copy_from=max_x_array, unit=stream.x.unit)
    result.channels[0] = ch.SciChannel(copy_from=max_y_array, unit=data_channel.unit)
    return result

def find_highest_peak(stream: ss.SciStream, channel_index: int = 0, **kwargs) -> tuple[bool, float, float]:
    """ Returns the highest peak in a data channel.
    
    Returns
    -------
    tuple : (bool, x, y)
        First boolean  indicates if a peak could be detected
        If True, then x, y is the coordinate of the peak
    """
    peaks = find_peaks(stream, channel_index, **kwargs)
    if peaks.get_stream_length() >= 1:
        max_indexes = np.where(peaks.channels[0].value == np.amax(peaks.channels[0].value))
        highest_peak_index = int(max_indexes[0])
        return (True, peaks.x.value[highest_peak_index], peaks.channels[0].value[highest_peak_index])
    else:
        return (False, 0, 0)

def calc_poly_fit(stream: ss.SciStream, channel_index: int = 0, degree: int = 1) -> np.ndarray:
    """ Calculates a polynomial fit of a data channel.
    It is a non exception throwing version of numpy's np.polyfit(). See detailed function description there
    
    Results
    -------
    fir_param : ndarray
        either the result of the poly_fit or an array with size = 0
    """
    try:
        fit_param = np.polyfit(stream.x.value, stream.channels[channel_index].value, deg=degree)
    except :
        fit_param = np.array([]) 
    return fit_param

class fft_window_type(enum.Enum): 
    uniform = enum.auto()   # no windowing (all coefficient are 1.0)
    hanning = enum.auto()   # windowing function useful for noise and single frequencies (amplitude most accurate but broader spectrum)
    hamming = enum.auto()   # windowing function useful for signals with many frequency components (amplitude not so accurate but higher resolution)
    blackman = enum.auto()

def calc_fft(data_samples: ch.SciChannel, samplerate: float, window: fft_window_type = fft_window_type.hanning, spectral_density: bool = False, compress: bool = False, **kwargs) -> ss.SciStream:
    """ calculate amplitude or spectral density spectrum from time data array(s) 
    
    Parameters
    ----------
    data_samples: SciChannel
        A array of data points sampled at equally time differences 
    samplerate: float
        defines how fast data points where measured. Provided in [Hz]   
    window: optional, fft_window_type
        defines the data windowing function to be used. Available functions are defined in enum fft_window_type 
    spectral_density: optional, bool
        if True then the spectrum is normalized by the frequency resolution and the resulting unit is 1/sqrt(Hz)
    compress: optional, bool
        if True, the spectrum data is compress. it uses "rms" compression if spectral_density is True otherwise "mean" 
    Result
    ------
        DataChannel
            Spectrum as SciStream  
    """
    # handle depreciated keyword for "spectral_density" 
    if "powerspectrum" in kwargs:
        spectral_density = kwargs["powerspectrum"]

    n_samples = data_samples.value.shape[0]

    #prepare result arrays--------------------------
    n_fft_points = int(n_samples/2+1)

    fft_result_array = np.zeros(n_fft_points)

    # define FFT windowing --------------------------------
    if window == fft_window_type.uniform:
        s_window_array = np.ones(n_samples)
        noise_power_bandwidth = 1.0

    elif window == fft_window_type.hanning:
        s_window_array = np.hanning(n_samples)/0.5
        noise_power_bandwidth = 1.5

    elif window == fft_window_type.hamming:
        s_window_array = np.hamming(n_samples)/0.54
        noise_power_bandwidth = 1.36

    elif window == fft_window_type.blackman:
        s_window_array = np.blackman(n_samples)
        noise_power_bandwidth = 1.73
    else:
        print(f"calc_frq_spectrum: Error: unknown windowing function selected: {window}")
        return ss.SciStream()
    
    # calculate FFT spectrums --------------------------------
    fft_data_array = np.fft.fft(s_window_array * data_samples.value)
    fft_amp_array = np.abs(fft_data_array[:n_fft_points]) / n_fft_points

    # convert to power spectrum if needed
    if spectral_density:
        fft_result_array = fft_amp_array / np.sqrt(samplerate / n_fft_points * noise_power_bandwidth)
    else:
        fft_result_array = fft_amp_array

    # create X-Axis vector with frequencies
    if samplerate > 0.0:
        fft_frequency_array = np.linspace(0.0, samplerate/2.0, n_fft_points, endpoint=True)
    else:
        fft_frequency_array = np.linspace(0.0, n_fft_points, 1.0, endpoint=False)

    # assemble the resulting FFT spectrum as SciStream
    result = ss.SciStream((fft_frequency_array, fft_result_array))
    result.x.unit = "Hz"
    if spectral_density:
        result.channels[0].unit = f"{data_samples.unit}{get_spectral_density_unit()}"
    else:
        result.channels[0].unit = data_samples.unit
    if compress:
        compress_method = compress_spec_algo.rms if spectral_density else compress_spec_algo.mean
        result = calc_compressed_spectrum(result, algo=compress_method)
    return result

class compress_spec_algo(enum.Enum): 
    none = enum.auto()  # no compression, return original stream
    max = enum.auto()   # calc maximum value of signal per compression slot
    mean = enum.auto()  # calc mean value of signal per compression slot 
    rms = enum.auto()   # calc rms value of signal per compression slot 
    custom = enum.auto()   # uses custom defined function to compress of signal 

def calc_compressed_spectrum(spec_data: ss.SciStream, min_dist_factor=1.02, algo:compress_spec_algo = compress_spec_algo.mean, custom_func = None) -> ss.SciStream:
    """Reduces number of data points in large data sets by logarithmic compression method
        uses a predefined or custom compression algorithm

    Parameters
    ----------
    spec_data: DataArray
        original spectrum to be compressed
    min_dist_factor
        defines the minimal distance of two frq_array points must have to compress these. Distance > frq_array[i]/frq_array[i-1] 
    algo: enum of type compress_spec_algo, optional
        if none, no compression happen at all, otherwise use one of the predefined algorithms or the supplied custom_func
    custom_func: function points, optional
        if algo is compress_spec_algo.custom, then use this function for calculating the pressed value 

    Result
    ------
        compressed_data_channel: SciStream
            new compressed array of frequency data points, not equally spaced anymore
    """
    def func_max(value_array: np.ndarray, from_index:int, to_index:int) -> float:
        return value_array[from_index:to_index].max()

    def func_mean(value_array: np.ndarray, from_index:int, to_index:int) -> float:
        return np.mean(value_array[from_index:to_index])

    def func_rms(value_array: np.ndarray, from_index:int, to_index:int) -> float:
        return np.sqrt(np.mean(np.square(value_array[from_index:to_index])))

    if algo == compress_spec_algo.none:
      result = spec_data
    else:    
        algo_map = {}
        algo_map[compress_spec_algo.mean] = func_mean
        algo_map[compress_spec_algo.max] = func_max
        algo_map[compress_spec_algo.rms] = func_rms
        algo_map[compress_spec_algo.custom] = custom_func

        compressed_frq:list[float] = spec_data.get_stream_range().value.tolist()[0:2]
        compressed_data:list[list[float]] = [ch.value.tolist()[0:2] for ch in spec_data.channels]

        compression_func = algo_map[algo]
        data_len = spec_data.get_stream_length()
        last_source_index=1 # last position in data, that was included in calculation of filter   
        for next_index in range(2,data_len):
            frq_data = spec_data.get_stream_range().value
            if frq_data[next_index]/frq_data[last_source_index] >= min_dist_factor:
                compressed_frq.append(stat.mean(frq_data[(last_source_index+1):(next_index+1)]))
                for ch_index in range(spec_data.get_channel_count()):
                    amp_data = spec_data.get_channel(ch_index).value
                    compressed_data[ch_index].append(compression_func(amp_data,last_source_index+1,next_index+1))
                last_source_index = next_index

        compressed_frq = ch.SciChannel(compressed_frq, unit=spec_data.get_stream_unit())

        result = ss.SciStream(compressed_frq, channels=spec_data.get_channel_count())
        for ch_index in range(spec_data.get_channel_count()):
            result.set_channel(ch_index, compressed_data[ch_index], unit=spec_data.get_channel_unit(ch_index))
    return result


def remove_peaks_in_spectrum(spec_data: ss.SciStream, channel_index: int = 0, frq_peaks: list[float] = [], span=1) -> ss.SciStream:
    """removes peaks at certain frequencys 
    
    Parameter
    ---------
    spec_data: SciStream
        The original spectrum
    channel_index: int
        The channel to smooth out
    frq_peaks: list[float]
        array of frequencys to be smoothed out
    span              
        defined as number of frequency index left and right of the peak to be used to smooth out the peak
    
    Returns
    -------
    smoothed_spec: SciStream
    """ 
    result = ss.SciStream(spec_data)
    for frq in frq_peaks:
        peak_ind = np.where(spec_data.x.value >= frq)[0][0]
        
        for smooth_index in range(peak_ind-span,peak_ind+span):
            result.channels[channel_index].value[smooth_index] = result.channels[channel_index].value[smooth_index+span*2+1]
    return result

def get_total_harmonic_distortion(spec_data: ss.SciStream, channel_index: int = 0, fundamental_frequency: float = 0.0, max_number_of_harmonics: int = 0) -> float:
    """ calculates the distortion of a signal by measuring its harmonics 
    
    Parameters
    ----------
    spec_data: SciStream
        The data source 
    channel_index: int
        The data source  channel to analyze
    fundamental_frequency
        the fundamental frequency from which the THD calculation shall be done. Defined in [Hz]. The nearest frq found in the frq_array is used 
    max_number_of_harmonics 
        optional, limits the number of harmonic frequency to used for THD calculation, 0 tells the function to use all possible harmonics found in the spectrum
    
    Result
    ------  
        thd - distortion in [dB]
    """ 
    ind = np.where(spec_data.x.value >= fundamental_frequency)[0][0]
    list_of_harmonics = range(ind*2, len(spec_data.x.value), ind)
    if max_number_of_harmonics > 0:
        list_of_harmonics = list_of_harmonics[:max_number_of_harmonics] # limit number of harmonics to be measured

    thd = 20.0 * math.log10(np.sqrt(sum(spec_data.channels[channel_index].value[list_of_harmonics]**2))/spec_data.channels[channel_index].value[ind])
    return thd

def get_noise_floor(spec_data: ss.SciStream, channel_index: int = 0, one_over_f_edge_frq: float = 0.0, level_of_percentile: float = 50.0, res_noise_floor_spec: ss.SciStream = None) -> float:
    """ determine the noise floor of a spectrum.
    
    With the level_of_percentile one can define the amount of spike and noise hills to be suppressed.
    With a value of 50 in most cases the background noise is returned. 
    With 100 maximal value found in the spectrum is returned
    With 80-95 small spikes (50Hz or so) can be ignored but real unexpected noise increase is detected

    Can be nicely combined with the function "calc_compressed_spectrum()" to improve result

    Parameter
    ---------
    spec_data: SciStream
        The data source 
    channel_index: int
        The data source  channel to analyze
    one_over_f_edge_frq:  optional
        if defined compensate for 1/f noise increase below edge frequency. Defined in [Hz]. 
    level_of_percentile:  optional
        defines the amount of data to be inside the noise_level. Useful to compensate for spikes. Has to be from 1-100 [%]. 
    res_noise_floor_spec: optional
        if defined the used spectrum for noise_floor calculation is returned. Useful to check for 1/f compensation
    
    Result
    ------  
    noise_level:
        noise level found 
    """ 
    # compensate for 1/f noise increase below edge frequency
    if one_over_f_edge_frq > 0.0:
        # # scale 1/f noise below frequency point
        low_data_array = spec_data.channels[channel_index].value[spec_data.x.value < one_over_f_edge_frq]
        low_frq_array  = spec_data.x.value[spec_data.x.value < one_over_f_edge_frq]
    
        low_data_array = low_data_array * np.sqrt(low_frq_array) / np.sqrt(one_over_f_edge_frq)
    
        low_data_array = low_data_array[2:] # remove DC
        low_frq_array  = low_frq_array[2:]
        
        res_data_array = np.concatenate((low_data_array, spec_data.channels[channel_index].value[len(low_data_array):]))
    else:
        res_data_array = spec_data.channels[channel_index].value

    noise_floor = np.percentile(res_data_array, level_of_percentile)

    if isinstance(res_noise_floor_spec, ss.SciStream):
        res_noise_floor_spec.channels[0].value = res_data_array
        res_noise_floor_spec.x.value = spec_data.x.value
    return noise_floor

def get_amplitude(spec_data: ss.SciStream, channel_index: int = 0, frq: float = 0.0) -> tuple[float, int]:
    """ returns the amplitude value of a specified frequency
        The amplitude value of the closest frequency found in the array is returned         
    
    Parameter
    ---------
    spec_data: SciStream
        The data source 
    channel_index: int
        The data source  channel to analyze
    frq : float
        the frequency of interest in [Hz]

    Result
    ------
    (amplitude, found_index): tuple(float, int)
        The amplitude of the data array at nearest frq
        found_index, the position index in data array. If not found, index is negativ
    """ 
    found_index = -1
    
    ind = np.where(spec_data.x.value >= frq)
    found_any = ind[0].shape[0]
    
    if found_any > 0:
        found_index = ind[0][0]
    
    if found_index >= 0:
        amp = spec_data.channels[channel_index].value[found_index]
    else:
        amp = 0.0
    return (amp, found_index)

def calc_signal_integral_of_frq_band(spec_data: ss.SciStream, start_frq: float, end_frq:float, spectral_density=False) -> float:
    """ calculates the area of data in a specified frequency band

    Parameters
    ----------
        spec_data: SciStream:
             with frequency data in x channel and signal data in channel 0
        start_frq, end_frq: float
            defines the start and end frequency of the desired frequency band
        spectral_density: bool
            if true, the amplitude data are assumed to have unit [1/sqrt(Hz)]
    
    Result
    ------
        amp_integral: float
            amplitude area from start_frq to end_frq in [amp_unit]
    Exception
    ---------
        ValueError: 
            In case the start or end_frq are not found in provided spec_data stream this exception is thrown
    """
    amp_integral = 0.0
    try:
        start_index: int = np.where(spec_data.x.value >= start_frq)[0][0]
        end_index: int = np.where(spec_data.x.value >= end_frq)[0][0]
    except:
        raise ValueError(f"Start or End value not found in spec_data.x stream: start_frq={start_frq}, end_frq={end_frq}")
    else:
        for i in range(start_index, end_index):
            current_value = spec_data.channels[0].value[i]
            current_frq_band = spec_data.x.value[i+1] - spec_data.x.value[i]
            if spectral_density:
                current_value = current_value*current_value
            amp_integral += current_value*current_frq_band
    if spectral_density:
        amp_integral = np.sqrt(amp_integral)
    return amp_integral

def calc_signal_integral_versus_upper_bandwidth(spec_data: ss.SciStream, start_frq: float, end_frq:float, spectral_density=False) -> ss.SciStream:
    """ calculates an array of signal amplitudes versus frequency upper bound with fix start frequency
        It creates the result array of amplitude integrals by sweeping from start_frq to end_frq and calculates for each step the signal integral.

    Parameters
    ----------
        spec_data: SciStream:
             with frequency data in x channel and signal data in channel 0
        start_frq, end_frq: float
            defines the start and end frequency of the desired frequency band
        spectral_density: bool
            if true, the amplitude data are assumed to have unit [1/sqrt(Hz)]
    
    Result
    ------
        SciStream: 
            amplitude array of signal amplitude versus frequency upper bound with fix start frequency
    Exception
    ---------
        ValueError: 
            In case the start or end_frq are not found in provided spec_data stream this exception is thrown
    """
    result = ss.SciStream()
    try:
        start_index: int = np.where(spec_data.x.value >= start_frq)[0][0]
        end_index: int = np.where(spec_data.x.value >= end_frq)[0][0]
    except:
        raise ValueError(f"Start or End value not found in spec_data.x stream: start_frq={start_frq}, end_frq={end_frq}")
    else:
        integral_value_array = np.zeros(end_index - start_index) 
        integral_frq_array = spec_data.x.value[start_index:end_index]

        for i in range(len(integral_value_array)):
            integral_frq_array[i] = spec_data.x.value[start_index + i]
            integral_value_array[i] = calc_signal_integral_of_frq_band(spec_data, start_frq, integral_frq_array[i], spectral_density)
    
        result = ss.SciStream((integral_frq_array, integral_value_array), x_unit=spec_data.x.unit)
        result.set_channel_unit(0, spec_data.get_channel_unit(0).removesuffix(get_spectral_density_unit()))
    return result


""" Application for performing Frequency Sweeps with Nanosurf Software
Compatible with Nanosurf software CX_3.10.3.5 and newer, Nanosurf python package 1.2.2 and DriveAFM
Author: Hans Gunstheimer
Copyright Nanosurf AG 2022
License - MIT
"""

import sys
import nanosurf.lib.datatypes.prop_val as prop_val

from app import app_base
from modules.frequency_sweep_module import sweep_module as sweep_func
from modules.frequency_sweep_module import sweep_gui as sweep_gui

MyCompany = "Nanosurf AG"
MyAppNameShort = "FrequencySweep"
MyAppNameLong = "Frequency Sweep Application"

class MyAppSettings(app_base.AppSettings):
    """ Settings defined here as PropVal are stored persistently in a ini-file"""
    AppHelloMsg = prop_val.PropVal("Welcome to the Nanosurf Frequency Sweep App")
    
class MyApp(app_base.ApplicationBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = MyAppSettings()

    def do_startup(self):
        """ Insert any module used in this app. 
            If more than one module is added, a menu bar is shown in the app window
        """
        self.add_module(sweep_func.FrequencySweepModule(self, sweep_gui.SweepScreen()), "Sweep Module")

        # here we apply a setting value just for fun
        self.show_message(self.settings.AppHelloMsg.value)
        
    def do_shutdown(self):
        """ Handle cleaning up stuff here"""
        pass

if __name__ == "__main__":
    App = MyApp(MyCompany, MyAppNameShort, MyAppNameLong, [])
    App.start_app()
    sys.exit(App.exec_())



""" Helper function to connect gui widgets to Property classes
Copyright Nanosurf AG 2021
License - MIT
"""
from enum import IntEnum
import pathlib
from PySide2 import QtWidgets
from nanosurf.lib.gui import nsf_sci_edit, nsf_edit, nsf_combobox
from nanosurf.lib.datatypes import sci_val
from nanosurf.lib.datatypes.prop_val import PropVal

def connect_to_property(widget: QtWidgets.QWidget, prop: PropVal) -> bool:
   """
   Connect the value of a property to a widgets. Keeps the widget content and the property value in sync
   Note: Not every QWidget can be connected to any Property type. 
   """
   is_supported = True
   if isinstance(widget, nsf_sci_edit.NSFSciEdit) and isinstance(prop.var, sci_val.SciVal):
      widget.set_value(prop.value)
      widget.set_unit(prop.var.unit())
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(prop.value)
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(widget.value())
      )
   elif isinstance(widget, nsf_sci_edit.NSFSciEdit) and isinstance(prop.var, float):
      widget.set_value(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(prop.value)
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(widget.value())
      )
   elif isinstance(widget, nsf_sci_edit.NSFSciEdit) and isinstance(prop.var, int):
      widget.set_value(prop.value)
      widget.set_precision(0)
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(prop.value)
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(int(widget.value()))
      )
   elif isinstance(widget, QtWidgets.QLineEdit) and isinstance(prop.var, str):
      widget.setText(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.setText(prop.value)
      )
      widget.textEdited.connect(lambda : 
         prop.set_value(widget.text())
      )
   elif isinstance(widget, QtWidgets.QLineEdit) and isinstance(prop.var, pathlib.Path):
      widget.setText(str(prop.value))
      prop.sig_value_changed.connect(lambda : 
         widget.setText(str(prop.value))
      )
      widget.textEdited.connect(lambda : 
         prop.set_value(pathlib.Path(widget.text()))
      )
   elif isinstance(widget, nsf_edit.NSFEdit) and isinstance(prop.var, pathlib.Path):
      widget.set_value(str(prop.value))
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(str(prop.value))
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(pathlib.Path(widget.value()))
      )
   elif isinstance(widget, nsf_edit.NSFEdit) and isinstance(prop.var, str):
      widget.set_value(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(prop.value)
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(widget.value())
      )
   elif isinstance(widget, QtWidgets.QCheckBox) and isinstance(prop.var, bool):
      widget.setChecked(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.setChecked(prop.value)
      )
      widget.stateChanged.connect(lambda : 
         prop.set_value(widget.checkState() > 0)
      )
   elif isinstance(widget, QtWidgets.QLabel):
      widget.setText(str(prop.value))
      prop.sig_value_changed.connect(lambda : 
         widget.setText(str(prop.value))
      )
   elif isinstance(widget, nsf_combobox.NSFComboBox) and isinstance(prop.var, int):
      widget.set_value(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.set_value(prop.value)
      )
      widget.value_changed_event.connect(lambda : 
         prop.set_value(widget.value())
      )
   elif isinstance(widget, QtWidgets.QComboBox) and isinstance(prop.var, int):
      widget.setCurrentIndex(prop.value)
      prop.sig_value_changed.connect(lambda : 
         widget.setCurrentIndex(prop.value)
      )
      widget.currentIndexChanged.connect(lambda : 
         prop.set_value(int(widget.currentIndex()))
      )
   else:
      is_supported = False
      raise TypeError(f"Error: Widget of type '{type(widget)}' cannot be connected to property of type '{type(prop.var)}'")
   return is_supported

   
   
  


"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import typing
import numpy as np
import nanosurf.lib.datatypes.sci_channel as ch

class SciStream:
    
    def __init__(self, 
        source: typing.Union[tuple, list[typing.Any, typing.Any], 
        ch.SciChannel, 'SciStream', None] = None, 
        channels: int = 1, stream_length: int = 0, 
        x_unit: str = ""):

        self.channels = [ch.SciChannel(array_length=stream_length) for d in range(channels)]
        self.x = ch.SciChannel(unit=x_unit)
        
        if isinstance(source, ch.SciChannel):
            self.x = ch.SciChannel(source)
            self.channels = [ch.SciChannel(np.zeros_like(source))]
        elif isinstance(source, SciStream):
            self.x = ch.SciChannel(source.x)
            self.copy_channels(source)
        elif source is not None:
            self.x = ch.SciChannel(source[0],unit=x_unit)
            self.channels = [ch.SciChannel(source[1])]

    def set_stream_length(self, length: int):
        if length != self.get_stream_length():
            if self.x.value.size > 0:       
                min = np.min(self.x.value)
                max = np.max(self.x.value)
            else:
                min = 0.0
                max = 1.0
            self.x.value = np.resize(self.x.value, length)
            self.define_stream_range(min, max)
            self._adjust_channel_length()

    def get_stream_length(self) -> int:
        return self.x.value.size

    def set_stream_range(self, range: typing.Union[ch.SciChannel, np.ndarray, list], unit: str = ""):
        if isinstance(range, ch.SciChannel):
            self.x = range
        else:
            self.x = ch.SciChannel(range, unit=unit)
        self._adjust_channel_length()

    def get_stream_range(self) -> ch.SciChannel:
        return self.x

    def define_stream_range(self, min: float, max: float, unit: str = ""):
        self.x.value = np.linspace(min, max, self.get_stream_length())
        if unit != "":
            self.x.unit = unit

    def get_stream_unit(self) -> str:
        return self.x.unit

    def set_stream_unit(self, unit: str):
        self.x.unit = unit

    def get_channel_count(self) -> int:
        return len(self.channels)
        
    def set_channel_count(self, channels: int):
        current_len = self.get_channel_count()
        if current_len < channels:
            self.channels.extend( [ch.SciChannel(array_length=self.get_stream_length()) for d in range(channels - current_len)])
        elif current_len > channels:
            self.channels = self.channels[:channels]

    def copy_channels(self, source: 'SciStream'):
        assert source.get_stream_length() == self.get_stream_length(), "SciStream: Streams must have equal length"
        self.channels = [ch.SciChannel(source.channels[stream_index]) for stream_index in range(source.get_channel_count())]    

    def get_channel(self, channel_index: int) -> ch.SciChannel:
        assert channel_index < len(self.channels), "SciStream: Channel index out of bound error"
        return self.channels[channel_index]

    def set_channel(self, channel_index: int, source: typing.Union[ch.SciChannel, np.ndarray, list], unit: str = ""):
        assert channel_index < len(self.channels), "SciStream: Channel index out of bound error"
        self.channels[channel_index] = ch.SciChannel(source, unit=unit)

    def get_channel_unit(self, channel_index: int) -> str:
        assert channel_index < len(self.channels), "SciStream: Channel index out of bound error"
        return self.channels[channel_index].unit

    def set_channel_unit(self, channel_index: int, unit: str):
        assert channel_index < len(self.channels), "SciStream: Channel index out of bound error"
        self.channels[channel_index].unit = unit

    def _adjust_channel_length(self):
        x_size = self.x.value.size
        for ch in self.channels:
            if ch.value.size != x_size:
                ch.value = np.resize(ch.value, x_size)


""" The long lasting worker thread which start the scan and store the scan lines
Copyright Nanosurf AG 2021
License - MIT
"""
import time
import pythoncom # to connect to SPM Controller within thread
from PySide2.QtCore import Signal

import nanosurf
from nanosurf.lib.datatypes import nsf_thread, sci_channel, sci_stream
from app import app_common, module_base
from modules.scan_module import settings

class ScanData():
    def __init__(self):
        self.stream = sci_stream.SciStream(channels=2)
        self.scan_line_index = -1

class ScanFrameWorker(nsf_thread.NSFBackgroundWorker):
    """ This class implements the long lasting activity of reading the scan data in the background to not freeze the gui """    

    sig_tick = Signal() # is emitted if par_send_tick is True
    sig_new_data = Signal() # is emitted each time a new scan line is measured. Result is read by self.get_result()

    """ parameter for the background work"""
    
    par_image_size = 1e-9
    par_time_per_line = 1.0
    par_points_per_line = 128
    par_channel_id = settings.ChannelD.Topography
    par_send_ticks = True
    par_time_between_ticks = 0.1

    _sig_message = Signal(str, int)

    def __init__(self, my_module: module_base.ModuleBase):
        """ setup the thread and wait until the task is started by thread.start()"""
        self.module = my_module
        self.resulting_data = ScanData()
        self.spm:nanosurf.spm.Spm = None
        super().__init__()
        self._sig_message.connect(self.module.app.show_message)

    def do_work(self):
        """ This is the working function for the long task"""
        self.resulting_data = ScanData()

        if self._connect_to_controller():
            self.spm_scan = self.spm.application.Scan

            #first stop a running scan
            if self.spm_scan.IsScanning:
                self.spm_scan.Stop()
                while self.spm_scan.IsScanning:
                    time.sleep(0.01)

            # prepare frame
            self.spm_scan.ImageSize(self.par_image_size, self.par_image_size)
            self.spm_scan.Scantime = self.par_time_per_line
            self.spm_scan.Points = self.par_points_per_line
            self.spm_scan.Lines = self.par_points_per_line
            
            self.resulting_data.stream.set_stream_length(self.par_points_per_line)
            self.resulting_data.stream.define_stream_range(min=0.0, max=self.par_image_size, unit="m")
            self.resulting_data.stream.set_channel_count(2)
            #start our frame
            self.spm_scan.StartFrameUp()

            # monitor scanning and read new scan lines as they are available
            self.next_scan_line = 0
            self.tick_time = 0.0
            while (self.next_scan_line < self.par_points_per_line) and not self.is_stop_request_pending():
                
                # wait for new data 
                self.tick_time = 0.0
                while (self.next_scan_line > (self.spm_scan.Currentline - 1)) and not self.is_stop_request_pending():
                    time.sleep(self.par_time_between_ticks)
                    self.tick_time += self.par_time_between_ticks
                    if self.par_send_ticks:
                        self.sig_tick.emit()
                    self.send_message(f"Wait for scan line {self.next_scan_line} since {self.tick_time:.2f}s")

                # read new measured line
                channel = self.par_channel_id
                self.resulting_data.stream.set_channel(0, self._read_scan_line(self.next_scan_line, channel, forward=True, raw=False))
                self.resulting_data.stream.set_channel(1, self._read_scan_line(self.next_scan_line, channel, forward=False, raw=False))
                self.resulting_data.scan_line_index = self.next_scan_line
                self.sig_new_data.emit()
                
                # prepare for next line
                self.next_scan_line += 1

            self.spm_scan.Stop()
            self.logger.info(f"Scan ended at line: {self.spm_scan.Currentline}")

        self._disconnect_from_controller()

    def get_result(self) -> ScanData:
        return self.resulting_data

    def send_message(self, msg:str, msg_type : app_common.MsgType = app_common.MsgType.Info):
        self._sig_message.emit(msg, msg_type)
        self.logger.info(msg)       
        
    def _read_scan_line(self, line: int, channel: int, forward=True, raw: bool = True) -> sci_channel.SciChannel:
        group  = nanosurf.Spm.ScanGroupID.Forward if forward else nanosurf.Spm.ScanGroupID.Backward
        filter = nanosurf.Spm.DataFilter.RAW if raw else nanosurf.Spm.DataFilter.LineFit
        unit_str = "m" if channel == 1 else "V"
        line_str:str = self.spm_scan.GetLine(group, channel, line, filter, nanosurf.Spm.DataConversion.Physical)
        line_str = line_str.replace(',','.') # needed in case the numbers are formatted with a ',' as decimal separator (e.g. standard german windows number format setting)
        line_split_str = line_str.split(";")
        scan_data = [float(p) for p in line_split_str]
        return sci_channel.SciChannel(scan_data, unit=unit_str)

    def _connect_to_controller(self) -> bool:
        self.send_message("Connecting to Nanosurf controller")
        pythoncom.CoInitialize()
        ok = False
        if self.spm is None:
            self.spm = nanosurf.SPM()
            if self.spm.is_connected():
                if self.spm.is_scripting_enabled():
                    ok = True
                else:
                    self.send_message("Error: Scripting interface is not enabled", app_common.MsgType.Error)
            else:
                self.send_message("Error: Could not connect to controller. Check if software is started", app_common.MsgType.Error)
        else:
            ok = True
        return ok

    def _disconnect_from_controller(self):
        if self.spm is not None:
            if self.spm.application is not None:
                del self.spm
            self.spm = None
        


"""Logical Unit Interface Runtime Wrapper
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import re
import pythoncom

def to_snake_case(word):
    # one could argue, that a number starts and ends a word, but this is
    # not too bad. (Numbers in variable names are a code smell anyhow)
    str1 = re.sub('([0-9]+)([A-Za-z]+)', r'\1_\2', word)
    str2 = re.sub('([A-Za-z])([0-9]+)', r'\1_\2', str1)
    str3 = re.sub('(^_)([A-Z][a-z]+)', r'\1_\2', str2)
    # break between lower case and upper case
    # this leaves abbreviations (consecutive capitals) together)
    return re.sub('([a-z])([A-Z])', r'\1_\2', str3).lower()

class _Attribute():
    """Accessor for LU Attributes"""
    # TODO: simplify the Attribute call
    # (it has too go too much back and forth to the parent).
    def __init__(self, parent, attribute_id):
        self._parent = parent
        self._id = attribute_id

    def __str__(self):
        try:
            value = self.value
        except pythoncom.com_error as exception:
            if ': Parameter error.' in exception.excepinfo[2]:
                value = "write-only"
        return f'{value} ({self.value_min} - {self.value_max}) {self.unit}'

    @property
    def value(self):
        return self._parent._logical_unit.Attribute(
            self._parent._type_id, self._parent._instance_id, self._id)

    @value.setter
    def value(self, value):
        self._parent._logical_unit.Attribute(
            self._parent._type_id, self._parent._instance_id, self._id, value)

    @property
    def unit(self):
        return self._parent._logical_unit.AttributeUnit(
            self._parent._type_id, self._parent._instance_id, self._id)

    @property
    def value_max(self):
        return self._parent._logical_unit.AttributeMax(
            self._parent._type_id, self._parent._instance_id, self._id)

    @property
    def value_min(self):
        return self._parent._logical_unit.AttributeMin(
            self._parent._type_id, self._parent._instance_id, self._id)


class _AttributeVector(_Attribute):
    class _ItemAccess():
        def __init__(self, parent, attribute_id):
            self._parent = parent
            self._id = attribute_id

        def __getitem__(self, index):
            return self._parent._logical_unit.AttributeVectorByIndexDouble(
                self._parent._type_id, self._parent._instance_id, self._id,
                index)

        def __setitem__(self, index, value):
            self._list[index] = value
            self._parent._logical_unit.AttributeVectorByIndexDouble(
                self._parent._type_id, self._parent._instance_id, self._id,
                index, value)

    def __init__(self, parent, attribute_id):
        super().__init__(parent, attribute_id)
        self._item_access = self._ItemAccess(
            parent, attribute_id)

    @property
    def item(self):
        return self._item_access

    @property
    def value(self):
        return self._parent._logical_unit.AttributeVectorDouble(
            self._parent._type_id, self._parent._instance_id, self._id)

    @value.setter
    def value(self, value):
        self._item_access._list = value
        self._parent._logical_unit.AttributeVectorDouble(
            self._parent._type_id, self._parent._instance_id, self._id, value)


class _LogicalUnit():
    """virtual base class for logical units."""
    _type_id = -1
    _attribute_dict = {}
    _logical_unit = None

    def __init__(self, instance_id=0):
        self._instance_id = instance_id

        for name, ident in self._attribute_dict.items():
            if to_snake_case(name)[-4:] == "_vec":
                setattr(
                    self, to_snake_case(name),
                    _AttributeVector(self, ident))
            else:
                setattr(
                    self, to_snake_case(name), _Attribute(self, ident))

    def __str__(self):
        result = ""
        for name, ident in self._attribute_dict.items():
            attr_name = to_snake_case(name)
            result = result + attr_name + ": " + str(getattr(self, attr_name)) + '\n'
        return result

    def _create_trigger(trigger_id):
        return lambda self: self._logical_unit.Trigger(
            self._type_id, self._instance_id, trigger_id)

    def _create_is_busy(busy_id):
        return property(lambda self: self._logical_unit.IsBusy(
            self._type_id, self._instance_id, busy_id))


def logical_unit_type(type_name, logical_unit, definition_dict):
    # The attributes must be created dynamically, because we
    # need to pass the self object, or some other properties of the
    # containing class to them.
    # I could find a way to do this with attributes created at class creation
    # time.
    type_dict = {
        '_type_id': definition_dict['id'],
        '_attribute_dict': definition_dict['attributes'],
        '_logical_unit': logical_unit,
        'Instance': enum.IntEnum('Instance', definition_dict['instances'])}

    for name, enum_dict in definition_dict['attribute_enumerations'].items():
        type_dict[name] = enum.IntEnum(name, enum_dict)

    # This could be done in the _LogicalUnit constructor as well, it is 
    # actually a bit strange that the capture of self by the lambda returned by
    # _create_trigger works at all.
    for name, ident in definition_dict['triggers'].items():
        type_dict[to_snake_case(name)] = _LogicalUnit._create_trigger(ident)

    # This must be called here, it cannot be called in the _LogicalUnit constructor.
    for name, ident in definition_dict['busy'].items():
        type_dict[to_snake_case(name)] = _LogicalUnit._create_is_busy(ident)

    return type(type_name, (_LogicalUnit,), type_dict)


if __name__ == "__main__":
    import manager_mock

    assert to_snake_case("CurrentAFMLockInYBinary") == "current_afm_lock_in_y_binary"

    test_lu_definition = {
        'id': 1,
        'instances': {'Z': 0, 'X': 1},
        'attributes':  {'input': 0, 'speed_vec': 1},
        'attribute_enumerations': {
            'Input':      {'x': 0, 'y': 1, 'z': 2},
            'Something':   {'NoThing': 0, 'OneThing': 1}
        },
        'triggers': {'start': 0, 'stop': 1},
        'busy': {'test_busy': 0, 'busy_too': 1}}
    test_lu3_definition = {
        'id': 3,
        'instances': {'BLI': 0, 'BLA': 1},
        'attributes':  {'output': 0, 'speedVec': 1},
        'attribute_enumerations': {
            'Output':      {'x': 0, 'y': 1, 'z': 2},
            'Something':   {'NoThing': 0, 'OneThing': 1}
        },
        'triggers': {'start': 0, 'stop': 1},
        'busy': {'test_busy': 0}}

    logical_unit = manager_mock.LogicalUnit()
    globals()['RampGen'] = logical_unit_type(
        'RampGen', logical_unit, test_lu_definition)
    globals()['TestLu2'] = logical_unit_type(
        'TestLu2', logical_unit, test_lu3_definition)

    ramp_z = RampGen(RampGen.Instance.Z)
    bli_unit = TestLu2(TestLu2.Instance.BLI)
    bla_unit = TestLu2(TestLu2.Instance.BLA)

    print("Test _logical_unit member")
    ramp_z._logical_unit.Attribute(1, 2, 3, 1.0)
    print(ramp_z._logical_unit.Attribute(1, 2, 3))
    print()

    print("Attribute directory:")
    print(dir(ramp_z))
    print()

    ramp_z.start()
    # Trig: Type_id: 1 Inst No: Instance.FIRST Trigger ID: 0
    ramp_z.stop()
    # Trig: Type_id: 1 Inst No: Instance.FIRST Trigger ID: 1
    print(ramp_z.test_busy)
    # False

    z_input = ramp_z.input
    z_input.value = RampGen.Input.z
    # SetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 0 Value: Input.z
    print(z_input.value)
    # GetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 0 Value: Input.z
    # Input.z
    print(z_input.value_max)
    # 10.0
    print(ramp_z.speed_vec.unit)
    # m  actually, all units will return m in this test.

    ramp_z.speed_vec.value = [0.1, 0.2]
    # SetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 1 Value: [0.1, 0.2]
    ramp_z.speed_vec.item[0] = 0.3
    # SetItem: Type ID: 1 Inst No: Instance.FIRST Attr ID: 1 At[0] Value: 0.3
    speeds = ramp_z.speed_vec.value
    # GetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 1 Value: [0.3, 0.2]
    print(speeds)
    # [0.3, 0.2]
    print(speeds[1])
    # 0.2
    print(ramp_z.speed_vec.item[0])
    # GetItem: Type ID: 1 Inst No: Instance.FIRST Attr ID: 1 At[0] Value: 0.3
    # 0.3

    print("\nNote that these statements have different effect:")
    z_input = ramp_z.input
    # -
    z_input_val = z_input.value
    # GetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 0 Value: Input.z
    print("")
    z_input.value = RampGen.Input.x
    # SetAttr: Type ID: 1 Inst No: Instance.FIRST Attr ID: 0 Value: Input.x
    z_input_val = RampGen.Input.y
    # -



""" Defines the configuration values of the module.
    Values in the ProStore are saved and loaded during application startup/shutdown automatically
Copyright Nanosurf AG 2021
License - MIT
"""

import enum
import nanosurf
import nanosurf.lib.datatypes.sci_val as sci_val
import nanosurf.lib.datatypes.prop_val as prop_val
import nanosurf.lib.spm.workflow.frequency_sweep as freq_sweep

class SweepSettings(prop_val.PropStore):
    """ settings defined here as PropVal are stored persistently in a ini-file
        settings with a '_' as first char are not stored
    """
    def __init__(self):
        self.cantilever = Cantilever()
        self.bandwidth = freq_sweep.Bandwidths.Hz_360
        self.input_source = freq_sweep.InputSource.Deflection
        self.output_source = freq_sweep.FrequencySweepOutput.Normal_Excitation
        
        self.center_frequency = prop_val.PropVal(sci_val.SciVal(5000, "Hz"))
        self.frequency_range = prop_val.PropVal(sci_val.SciVal(200, "Hz"))
        self.frequency_steps = prop_val.PropVal(sci_val.SciVal(10, "Hz"))
        self.excitation_amplitude = prop_val.PropVal(sci_val.SciVal(0.2, "V"))
        self.deflection_setpoint = prop_val.PropVal(sci_val.SciVal(0, "V"))

        self.plot_style_id = prop_val.PropVal(int(PlotStyleID.Linear))
        self.excitation_method = prop_val.PropVal(int(ExcitationMethodID.CleanDrive))

class SweepResults():
    """ This class saves the worker task result (e.g be read by gui elements or saved to file """
    def __init__(self):
        self.resonance_frequency = 0.0
        self.q_factor = 0.0
        self.spring_constant = 0.0

class PlotStyleID(enum.IntEnum):
    Linear = 0,
    Logarithmic = 1,

class ExcitationMethodID(enum.IntEnum):
    CleanDrive = 0,
    PiezoDrive = 1,

class Cantilever():
    def __init__(self):
        self.index = 0
        self.name = ""
        self.length = prop_val.PropVal(sci_val.SciVal(0, "m"))
        self.width = prop_val.PropVal(sci_val.SciVal(0, "m"))
        self.spring_constant = prop_val.PropVal(sci_val.SciVal(0, "N/m"))
        self.resonance_frequency_air = prop_val.PropVal(sci_val.SciVal(0, "Hz"))
        self.q_factor_air = prop_val.PropVal(sci_val.SciVal(0, ""))
        self.resonance_frequency_liquid = prop_val.PropVal(sci_val.SciVal(0, "Hz"))
        self.q_factor_liquid = prop_val.PropVal(sci_val.SciVal(0, ""))
        self.shape = ""
        self.tip_half_angle = prop_val.PropVal(sci_val.SciVal(0, "deg"))
        self.tip_radius = prop_val.PropVal(sci_val.SciVal(0, "m"))


"""Provide support for Nanosurf Accessory Interface Electronics
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

from array import array
import nanosurf.lib.spm.com_proxy as spm
import nanosurf.lib.devices.i2c.bus_master as i2c

class AISlaveIDHeader:
    """ This class stores the information found in the id eeprom of each slave device"""

    def __init__(self, binaryarray: list = [int]):
        """ This class stores the information found in the id eeprom of each slave device

            As standart identification a bt-number is read from the device
            Optional a serial number is provided

        Parameters
        ----------
        binaryarray : array of byte data, optional
            data array (e.g. read from eeprom) to be decoded and filled into the class structure
        """

        self._version = 0
        self._bt_number = ""
        self._serial_no = ""
        self.decode_eeprom_data(binaryarray)

    def decode_eeprom_data(self, binaryarray: list[int]):
        """ This function can extract the id header information stored in a binary data stream, normally read from a id eeprom.
            The information is stored as class variables and can be ready by accessor functions

        Parameters
        ----------
        binaryarray : array of byte data, optional
            data array (e.g. read from eeprom) to be decoded and filled into the class structure

       """
        if len(binaryarray) > 10:
            self._version = binaryarray[0]
            assert self._version == 1, "AISlaveIDHeader: unknown id header version detected."
            self._bt_number = array('B', binaryarray[2:2+binaryarray[1]]).tobytes().decode()
            if binaryarray[9] > 0:
                self._serial_no = array('B', binaryarray[10:10+binaryarray[9]]).tobytes().decode()
            else:
                self._serial_no = ""

    def get_bt_number(self) -> str:
        """ Returns the slave devices identification string.
            It's usual form is "BTxxxxx"

        Returns
        -------
        bt_number: str
            identification string

        """
        return self._bt_number

    def get_serial_number(self) -> str:
        """ Return the slave devices (optional) serial number.
            It's usual form is "xxx-xx-xxx"

         Returns
         -------
            serial_number: str
                 serial number as string or empty string if no serial number is defined

        """
        return self._serial_no

class AccessoryInterface(i2c.I2CBusMaster):
    """ This is the main class to get access to an accessory interface (AI) and its slave devices.

        Connect to a know AI by its serial-no, or scan first for available AIs, or connect to first found
        Then access to a slave device of this AI can be etablished by select_port()

        slave devices are identified by get_slave_device_id().
        To talk to a slave device, specific classes have to be build.

    """
    def __init__(self, spm: spm.Spm):
        """ This is the main class to get access to an accessory interface (AI) and its slave devices.

        Parameters
        ----------
        spm
            reference to the connected spm COM class

        """
        super().__init__(spm_ctrl=spm, bus_id=i2c.I2CBusID.Unassigned, 
            instance_id=i2c.I2CInstances.MAIN_APP, 
            master_type=i2c.I2CMasterType.ACCESSORY_MASTER, 
            bus_speed=i2c.I2CBusSpeed.kHz_200)
        self._own_id_header = AISlaveIDHeader()
        self._active_port = -1

    def get_list_of_available_interfaces(self) -> list[str]:
        """ Searches for attached accessory interfaces and create list of available devices
            Have to be called at least once before a connect() can be done

         Returns
         ------
         list of str
            list of serial numbers of all found accessory interfaces

        """
        self._list_of_found_ai = []

        count_ai = self._test_obj.GetAccessoryInterfaceInUseCount
        for i in range(count_ai):
            self._list_of_found_ai.append(self._test_obj.GetAccessoryInterfaceInUseSerial(i))

        count_ai = self._test_obj.GetAccessoryInterfaceAvailableCount
        for i in range(count_ai):
            self._list_of_found_ai.append(self._test_obj.GetAccessoryInterfaceAvailableSerial(i))

        return self._list_of_found_ai

    def connect(self, serial_no: str = "") -> bool:
        """ connect this instance to a accessory interface identified by device serial-number

        Parameters
        ----------
        serial_no : str, optional
            serial-number of accessory interface to be connected to. If omitted it connects to fist AI found

        Returns
         ------
        Bool
            true if connection could be set up

        """
        self._bus_id = i2c.I2CBusID.Unassigned # no connection
        self._active_port = -1

        # without given serial number we try auto detection
        if serial_no == "":
            self.get_list_of_available_interfaces()
            if len(self._list_of_found_ai) > 0:
                serial_no = self._list_of_found_ai[0]

        # setup connection
        try:
            self._bus_id = self.get_bus_addr_of_ai_device(serial_no)

            # if connection could be done read and store device information
            if self._bus_id != i2c.I2CBusID.Unassigned:
                self.select_port(0) # port 0 is the internal port with self identification eeprom
                self._own_id_header = self.get_slave_device_id()
            else:
                self._own_id_header = AISlaveIDHeader()

            self.select_port(1) # default port

        except AssertionError:
            self._own_id_header = AISlaveIDHeader()
            self._bus_id = i2c.I2CBusID.Unassigned

        self.assign_i2c_bus(self._bus_id, self._bus_speed)
        return self._bus_id != i2c.I2CBusID.Unassigned

    def get_port_count(self) -> int:
        """ return the number of slave port the connected interface has

        Returns
        -------
        int
            number of ports avaliable on this accessory interface

        """
        known_ai_devices = {"BT07172": 5}
        if self._own_id_header.get_bt_number() in known_ai_devices:
            return known_ai_devices[self._own_id_header.get_bt_number()]
        else:
            return 0

    def get_serial_number(self) -> str:
        """ returns the serial number of the connected accessory interface

        Returns
        -------
        str
            Onw serial_number as a string in the form "xxx-yy-zzz" or empty string if not connected

        """
        return self._own_id_header.get_serial_number()

    def get_bus_addr(self) -> int:
        """ returns the communication bus address. Used for further I2C communication with a slave

        Returns
        -------
        int
            bus_address - identification number to be used for I2C communication

        """
        return self._bus_id

    def select_port(self, port_nr: int):
        """ opens the port to communication with a slave device at port
            Only one port at a given time can be selected. all further slave communication goes through the selected port
            Port 0 is assigned to accessory interface internal configuration and should be used carefully

        Parameters
        ----------
        port_nr : int
            identification number of the port to be used
        """
        assert self._bus_id != i2c.I2CBusID.Unassigned, "Accessory Interface: Error: Slave Access, but connection not assigned"
        ret = self._test_obj.I2CSetupMetaData(0,1,self._bus_id,0x70,0,1) # Setup for switch(multiplexer)
        assert ret == 0, "Accessory Interface: Error: Cannot connect to bus switch"
        ret = self._test_obj.I2CIsConnected
        assert ret == 1,"Accessory Interface: Error: Bus switch is not responding"

        # activate port, port 0 is internal, port 1 to 5 are accessory ports
        ret = self._test_obj.I2CWrite(0,1,[1 << port_nr])
        assert ret == 0, "Accessory Interface: Error: Write error to Bus switch"
        self._active_port = port_nr

    def is_slave_device_connected(self) -> bool:
        """ check if a slave device is connected to selected port
            It looks if the id eeprom can be found

        Returns
        -------
        bool
            returns True if a device is found on selected port, otherwise False

        """

        assert self._bus_id != i2c.I2CBusID.Unassigned, "Accessory Interface: Error: Slave Access, but connection not assigned"
        ret = self._test_obj.I2CSetupMetaData(0,1,self._bus_id,0x57,2,1) # Setup for eeprom
        assert ret == 0, "Accessory Interface: Error: Cannot setup device metadata"
        id_eeprom_connected = self._test_obj.I2CIsConnected
        return id_eeprom_connected != 0

    def get_slave_device_id(self) -> AISlaveIDHeader:
        """ read the identification information from the device on selected port
            The information is read from the id eeprom and stored in a AISlaveIDHeader class

         input:
            none
         return:
            slave_id - AISlaveIDHeader class type with the retrieved information
        """
        assert self._bus_id != i2c.I2CBusID.Unassigned, "Accessory Interface: Error: Slave Access, but connection not assigned"
        ret = self._test_obj.I2CSetupMetaData(0,1,self._bus_id,0x57,2,1) # Setup for eeprom
        assert ret == 0, "Accessory Interface: Error: Cannot setup device metadata"
        id_eeprom_data = self._test_obj.I2CReadEx(0, 19)
        return AISlaveIDHeader(id_eeprom_data)

    def select_port_with_slave(self, serial_nr: str) -> bool:
        """ Auto select the port where a device with 'serial_nr' is connected

            Parameters:
            -----------
            serial_nr: str
                The serial number of the connected device to find. Need the form 'xxx-yy-zzz'

            returns True if found othervise False
        """
        device_found = False
        try:
            for port_nr in range(1, self.get_port_count()+1):
                self.select_port(port_nr)
                if self.is_slave_device_connected():
                    slave_id = self.get_slave_device_id()
                    if serial_nr == slave_id.get_serial_number():
                        device_found = True
                        break
        except AssertionError:
            pass
        return device_found

    def get_current_port(self) -> int:
        return self._active_port

    def assign_chip(self, chip: 'I2CChip', port: int = -1):
        """ Tell the software that a certain 'chip' is connected to current or specific port
        
        Parameters:
        -----------
        chip: class of I2CChip 
        port: int  
            if no port number is given (or -1) then  assume the chip is connected to current selected port
        """
        chip.assigned_ai_port = port if port>=0 else self._active_port
        super().assign_chip(chip)

    def get_bus_addr_of_ai_device(self, serial_no: str) -> int:
        """ returns the bus_address to a accessory interface with provided serial_no

        Parameters
        ----------
        serial_no : str
            serial-number of accessory interface to provide the bus_adr

        Returns
         ------
        bus_adr: int
            bus_adr of AI with provided serial_no or a negative number if AI was not found

        """

        found_bus_addr = i2c.I2CBusID.Unassigned # init return value

        # search for bus address of given serial number in list of InUse devices
        count_ai = self._test_obj.GetAccessoryInterfaceInUseCount
        for i in range(count_ai):
            if serial_no == self._test_obj.GetAccessoryInterfaceInUseSerial(i):
                found_bus_addr = self._test_obj.GetAccessoryInterfaceInUseI2CBus(i)
                break

        # if not found, search in list of free devices
        if found_bus_addr == i2c.I2CBusID.Unassigned:
            count_ai = self._test_obj.GetAccessoryInterfaceAvailableCount
            for i in range(count_ai):
                if serial_no == self._test_obj.GetAccessoryInterfaceAvailableSerial(i):
                    found_bus_addr = self._test_obj.ActivateAccessoryInterfaceAvailable(i)
                    break

        return found_bus_addr

    # overwrite base functions to support automatic port switching-----------------------------------------

    def activate_chip(self, chip: 'I2CChip'):
        if  chip.assigned_ai_port != self._active_port:
            self.select_port(chip.assigned_ai_port)
        super().activate_chip(chip)




""" defines some nanosurf colors
Copyright Nanosurf AG 2021
License - MIT
"""
class NSFColorHexStr():
    Orange = "EA390D" 
    Soft_Gray = "555555"
    Soft_White = "F0F0F0"

class NSFColorsTuple():
    Orange = (234,57,13)    
    Soft_Gray = (85,85,85)
    Soft_White = (240,240,240)


"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits

class _CtrlUnitADC(ctrlunits._CtrlUnit):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm)
        self._lu_inst_name = lu_inst_name

    @property
    def inst_name(self):
        return self._lu_inst_name

    @property
    def dc(self):
        return self._lu.current_input_value.value

    @property
    def dc_norm(self):
        return self._lu.current_input_value.value / self.max

    @property
    def dc_v(self):
        return self.dc_norm * self._voltage_range

    @property
    def unit(self):
        return self._lu.current_input_value.unit

    @property
    def max(self):
        return self._lu.current_input_value.value_max

    @property
    def inverted(self):
        return self._lu.calib_polarity.value == self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Negative

    @inverted.setter
    def inverted(self, val):
        if val == True:
            self._lu.calib_polarity.value = self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Negative
        else:
            self._lu.calib_polarity.value = self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Positive

    @property
    def calib_gain(self):
        return self._lu.calib_gain.value

    @calib_gain.setter
    def calib_gain(self, val):
        self._lu.calib_gain.value = val

    @property
    def calib_offset(self):
        return self._lu.calib_offset.value

    @calib_offset.setter
    def calib_offset(self, val):
        self._lu.calib_offset.value = val

class CtrlUnitADCHiRes(_CtrlUnitADC):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm,lu_inst_name)
        self._lu = self._spm.lowlevel.AnalogHiResIn(spm.lowlevel.AnalogHiResIn.Instance[lu_inst_name])
        if 'USER' in lu_inst_name:
            self._voltage_range = 10.0
        else:
            self._voltage_range = 5.0

class CtrlUnitADCFast(_CtrlUnitADC):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm,lu_inst_name)
        self._lu = self._spm.lowlevel.AnalogFastIn(spm.lowlevel.AnalogFastIn.Instance[lu_inst_name])
        self._voltage_range = 1.0




""" The GUI of the application framework
Copyright Nanosurf AG 2021
License - MIT
"""

from PySide2 import QtWidgets
from PySide2 import QtGui, QtWidgets, QtCore
import nanosurf as nsf
from app import app_common, module_base

control_layout_width = 200

class MenuSeparator(QtWidgets.QFrame):
    def __init__(self, hidden : bool = False, height: int = 1,  **kargs):
        super().__init__(*kargs)
        self.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Plain)
        self.setStyleSheet(f"background-color:#{nsf.gui.nsf_colors.NSFColorHexStr.Orange};")
        self.setFixedHeight(height)
        self.setHidden(hidden)

class MenuButton(QtWidgets.QWidget):
  
    sig_on_menu_clicked = QtCore.Signal(int)
   
    def __init__(self, menutext: str, menuitem: int):
        super().__init__()
        self._setup_widgets(menutext)
        self.set_highlight(False)
        self.menuitem = menuitem
        self._button.clicked.connect(self._on_clicked)
    
    def set_highlight(self, highlight: bool = False):
        color = nsf.gui.nsf_colors.NSFColorHexStr.Orange if highlight else nsf.gui.nsf_colors.NSFColorHexStr.Soft_Gray
        self._marker_line.setStyleSheet(f"background-color:#{color};")

    def _on_clicked(self):
        self.sig_on_menu_clicked.emit(self.menuitem)

    def _setup_widgets(self, label_str: str):
        layout = QtWidgets.QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0,0,0,0)
        self._button = QtWidgets.QPushButton(label_str)
        self._marker_line = MenuSeparator(hidden=False, height=2)
        layout.addWidget(self._marker_line)
        layout.addWidget(self._button)
        self.setLayout(layout)      


class StdVSpacer(QtWidgets.QSpacerItem):
    def __init__(self):
        super().__init__(control_layout_width, 40, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.MinimumExpanding)

class AppWindow(QtWidgets.QMainWindow):
    def __init__(self, app):
        super().__init__()
        self.app = app

    def create_gui(self, ui_style_file : str, icon_file : str):
        # create main GUI layout
        style_sheet_str = self._read_style_sheet_into_str(ui_style_file)
        self.setStyleSheet(style_sheet_str)
        self.setWindowTitle(self.app.app_name_long)
        self.setWindowIcon(QtGui.QIcon(str(icon_file)))
        self.statusBar().showMessage("Ready")

        # setup general GUI layout
        # on the left will be a menu to open the modules stacked guis on the right
        self.mainScreen = QtWidgets.QWidget()
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_menu = QtWidgets.QHBoxLayout()
        self.main_messagebox = nsf.gui.NSFInfoBox(hidden=True)
        self.main_stack = QtWidgets.QStackedLayout()
        self.main_stack.addWidget(QtWidgets.QLabel("Main screen"))

        self.menu_sep = MenuSeparator(hidden=True)
        self.status_sep = MenuSeparator(hidden=False)
        self.main_layout.addLayout(self.main_menu,0)
        self.main_layout.addWidget(self.menu_sep,1)
        self.main_layout.addWidget(self.main_messagebox,2)
        self.main_layout.addLayout(self.main_stack,3)
        self.main_layout.addWidget(self.status_sep,4)
        #self.main_layout.addSpacerItem(StdVSpacer())
        self.mainScreen.setLayout(self.main_layout)
        self.setCentralWidget(self.mainScreen)
        self.load_window_size()

    def show_message(self, msg: str, msg_type: app_common.MsgType = app_common.MsgType.Info):
        if msg_type == app_common.MsgType.Error:
            self.show_info_box(msg, background_color=nsf.gui.nsf_colors.NSFColorHexStr.Orange, text_color="")
        elif msg_type == app_common.MsgType.Warn:
            self.show_info_box(msg, '')
        else:
            self.statusBar().showMessage(msg)        

    def save_window_size(self):
        self.app.registry.setValue("geometry", self.saveGeometry())
        self.app.registry.setValue("windowState", self.saveState())
    
    def load_window_size(self):
        self.restoreGeometry(self.app.registry.value("geometry"))
        self.restoreState(self.app.registry.value("windowState"))

    def add_screen(self, module: module_base.ModuleBase):
        module_gui = module.gui
        if isinstance(module_gui, QtWidgets.QWidget):
            self.main_stack.addWidget(module_gui)
            # activate first screen 
            if self.main_stack.currentIndex() == 0:
                self.main_stack.setCurrentWidget(module_gui)

    def get_active_module_index(self) -> int:
        return self.main_stack.currentIndex() - 1

    def set_active_module_by_index(self, index: int):
        self.main_stack.setCurrentIndex(index+1)
        self._update_menu_button_highlight()

    def is_menu_visible(self) -> bool:
        return self.app.get_module_count() > 1

    def update_menu(self, modules: dict):
        mod_with_gui = 0
        for mod in modules.values():
            if mod.gui is not None:
                mod_with_gui += 1

        # show menu only if there are more than one modules with gui present
        if mod_with_gui > 1:
            #remove all menu items
            while self.main_menu.count():
                child = self.main_menu.takeAt(0)
                if child.widget() is not None:
                    child.widget().deleteLater()

            # add menu items
            menuindex = 0
            for mod in modules.values():
                if mod.gui is not None:
                    menuitem = MenuButton(mod.name, menuindex)
                    menuitem.sig_on_menu_clicked.connect(self._on_menu_button_clicked)
                    self.main_menu.addWidget(menuitem)
                    menuindex += 1
            self.main_menu.addStretch() 

        self.menu_sep.setHidden(mod_with_gui <= 1)
        
    # internal use only

    def closeEvent(self, arg):
        """ capture close button from application window"""
        self.save_window_size()
        super().closeEvent(arg)

    def _read_style_sheet_into_str(self, style_file: str):
        style = ""
        with open(style_file) as f:
            style = f.read()
        return style
    
    def _on_menu_button_clicked(self, index: int):
        self.app.activate_module(index)

    def _update_menu_button_highlight(self):
        if self.is_menu_visible():
            active_index = self.get_active_module_index()
            for i in range(self.main_menu.count()-1):
                self.main_menu.itemAt(i).widget().set_highlight(i == active_index)

    def show_info_box(self, msg: str, background_color: str, text_color: str):
        self.main_messagebox.set_background_color(background_color)
        self.main_messagebox.set_text_color(text_color)
        self.main_messagebox.set_message(msg)
        self.main_messagebox.setHidden(False)



"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import time
import numpy as np
import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits

class _CtrlUnitSampler(ctrlunits._CtrlUnit):
    def __init__(self, spm):
        super().__init__(spm)
        self._lu = self._spm.lowlevel.DataAcquisition()
        self._databuffer = self._spm.lowlevel.DataBuffer()
        self._channel_list = []
        self._channel_index = {}
        self.data = {}
        self.data_max = {}
        self.data_unit = {}
        self._target_sampling_rate = self.sample_rate

    @property
    def samples(self):
        return self.__samples

    @samples.setter
    def samples(self, val):
        max_samples = self.samples_max
        if val > int(max_samples):
            val = int(max_samples)
        self.__samples = val

    @property
    def samples_max(self):
        max_samples = int(self._lu.sampler_datapoints.value_max)
        if max_samples > self._databuffer.available_points:
            max_samples = self._databuffer.available_points
        return int(max_samples)

    @property
    def channels(self):
        return self._channel_list

    @property
    def sample_rate(self):
        return self._lu.sampler_data_rate.value

    @sample_rate.setter
    def sample_rate(self, val):
        self._target_sampling_rate = val
        self._lu.sampler_data_rate.value = self._target_sampling_rate
        
    @channels.setter
    def channels(self, val):
        self._channel_list = val
        self._channel_index.clear()
        index = 0
        for ch in self._channel_list:
            self._channel_index[ch] = index
            index += 1

    def measure(self, setuptime: float = 2.0):
        self._lu.number_of_sampler_sets.value = 1
        self._lu.active_sampler_set.value = 0
        self._lu.phase_mode_enabled.value = 0
        self._lu.sampler_group_id.value = self._databuffer.group_id
        self._lu.sampler_datapoints.value = self.samples
        self._lu.sampler_auto_set_filter_mode.value = 1.0
        self._setup_inputs_and_mask()
        self._lu.sampler_trigger_mode.value = self._lu.SamplerTriggerModeID.InternalTimer
        self._lu.sampler_data_rate.value  = self._target_sampling_rate

        print("Measuring", end='')
        self._lu.sampler_start()
        time.sleep(setuptime) # these waiting time is important to let controller sw prepare the start and avoid wrong valid buffer message
        tick = 0
        while not self._databuffer.is_valid:
            time.sleep(0.01)
            if tick > 100:
                print(".", end='')
                tick = 0
            tick += 1
        print("")


        print("Read data", end='')
        self.data.clear()
        self.data_max.clear()
        self.data_unit.clear()
        for ch in self._channel_list:
            print(".", end='')
            data_channel = self._databuffer.channel(self._channel_index[ch])
            self.data[ch] = np.array(data_channel.data)/(pow(2, 31) - 1) * data_channel.dimension(2).data_range / 2.0
            self.data_max[ch] = data_channel.dimension(2).data_range / 2.0
            self.data_unit[ch] = data_channel.dimension(2).unit
        print("")

        self.timeline = np.linspace(0, 1.0 / self.sample_rate * (self.samples - 1), self.samples, endpoint=True)

    def _setup_inputs_and_mask(self):
        raise NotImplementedError

class CtrlUnitSamplerHiRes(_CtrlUnitSampler):
    def __init__(self, spm):
        super().__init__(spm)

    def _set_daq_input_channel(self, ch):
        index = self._channel_index[ch]
        if index == 0:
            self._lu.sampler_hi_res_ch_0_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 1:
            self._lu.sampler_hi_res_ch_1_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 2:
            self._lu.sampler_hi_res_ch_2_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 3:
            self._lu.sampler_hi_res_ch_3_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 4:
            self._lu.sampler_hi_res_ch_4_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 5:
            self._lu.sampler_hi_res_ch_5_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 6:
            self._lu.sampler_hi_res_ch_6_input.value = self._lu.SamplerHiResInSignal[ch]
        elif index == 7:
            self._lu.sampler_hi_res_ch_7_input.value = self._lu.SamplerHiResInSignal[ch]
        else:
            assert False, f"index={index} is out of range"

    def _setup_inputs_and_mask(self):
        mask = 0
        for ch in self._channel_list:
            mask |= 1 << self._channel_index[ch]
            self._set_daq_input_channel(ch)
        self._lu.sampler_channel_mask.value = mask

        




"""Package for scripting the Nanosurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import nanosurf.lib.spm.com_proxy as spm
from nanosurf.lib.spm.com_proxy import Spm
import nanosurf.lib.spm.lowlevel as ll
import nanosurf.lib.spm.lowlevel.ctrlunits.factory as ctrl_factory

import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits
import nanosurf.lib.spm.workflow as workflow
import nanosurf.lib.spm.scanhead as scanhead

from nanosurf.lib.spm.lowlevel.ctrlunits.capture import _CtrlUnitCapture
from nanosurf.lib.spm.lowlevel.ctrlunits.sampler import _CtrlUnitSampler
from nanosurf.lib.spm.lowlevel.ctrlunits.adc import _CtrlUnitADC
from nanosurf.lib.spm.lowlevel.ctrlunits.dac import _CtrlUnitDAC
from nanosurf.lib.spm.lowlevel.ctrlunits.analyzer import CtrlUnitsSineWaveGenerator, CtrlUnitLockIn
from nanosurf.lib.spm.lowlevel.ctrlunits.channelmux import _CtrlUnitChannelMux





"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import enum
import nanosurf.lib.spm.com_proxy as spm

class I2CBusID(enum.IntEnum):
    Unassigned = -1
    User      = 0x2000300
    HV        = 0x2000320
    ScanHead  = 0x2000360
    Interface = 0x2000340

class I2CInstances(enum.IntEnum):
    MAIN_APP = 0
    CONTROLLER = 1

class I2CMasterType(enum.IntEnum):
    AUTO_DETECT = -1
    EMBEDDED_AVALON = 0
    ACCESSORY_MASTER = 1
    EMBEDDED_LINUX = 2

class I2CBusSpeed(enum.IntEnum):
    kHz_Default = 0
    kHz_100 = 1
    kHz_200 = 2
    kHz_400 = 3

class I2CSyncing(enum.IntEnum):
    NoSync = 0
    Sync = 1
    
class I2CByteMode(enum.IntEnum):
    SingleByteOff = 0
    SingleByteOn = 1

class I2COffsetMode(enum.IntEnum):
    NoOffset = 0
    U8Bit = 1
    U16Bit_MSBFiRST = 2
    U16Bit_LSBFiRST = 3

class I2CBusMaster():
        
    _active_chip_ref = int(0)
    _next_chip_ref = 1

    def __init__(self, spm_ctrl: spm.Spm, bus_id: I2CBusID, instance_id: I2CInstances = I2CInstances.CONTROLLER, master_type: I2CMasterType = I2CMasterType.AUTO_DETECT, bus_speed: I2CBusSpeed = I2CBusSpeed.kHz_400):
        self.spm_ctrl = spm_ctrl
        self._test_obj = self.spm_ctrl.application.CreateTestObj
        self._rx_packet_buffer_len = 50
        self._tx_packet_buffer_len = 50
        self._instance_id = instance_id
        self._master_type = master_type 
        self._bus_id = bus_id
        self._bus_speed = bus_speed
        if self._master_type == I2CMasterType.AUTO_DETECT:
            self._master_type = self._auto_set_bus_master()

    def assign_i2c_bus(self,  bus_id: I2CBusID, bus_speed: I2CBusSpeed):
        self._bus_id = bus_id
        self._bus_speed = bus_speed

    def assign_chip(self, chip: 'I2CChip'):
        chip.setup_bus_connection(self, self.create_unique_chip_id())

    def setup_metadata(self, addr: int, offset_mode: I2COffsetMode, auto_lock: bool = True):
        self._test_obj.I2CSetupMetaDataEx(self._rx_packet_buffer_len, self._tx_packet_buffer_len, I2CSyncing.NoSync, I2CByteMode.SingleByteOff, self._bus_speed)
        self._test_obj.I2CSetupMetaData(self._instance_id, self._master_type, self._bus_id, addr, offset_mode, auto_lock)

    def check_connection(self, chip: 'I2CChip') -> bool:
        self.activate_chip(chip)
        is_connected:int = self._test_obj.I2CIsConnected
        return is_connected > 0

    @classmethod
    def get_active_chip_id(cls) -> int:
        return I2CBusMaster._active_chip_ref

    @classmethod
    def create_unique_chip_id(cls) -> int:
        I2CBusMaster._next_chip_ref += 1
        return I2CBusMaster._next_chip_ref

    def activate_chip(self, chip: 'I2CChip'):
        if chip.get_chip_ref() != I2CBusMaster._active_chip_ref:
            chip.activate()
            I2CBusMaster._active_chip_ref = chip.get_chip_ref()

    def write_bytes(self, offset: int, data: list[int]) -> bool:
        done = False
        try:
            done: bool = self._test_obj.I2CWrite(offset, len(data),data)
        except:
            pass
        return done

    def read_bytes(self, offset:int, data_count:int) -> list[int]:
        try:
            data: list[int] = self._test_obj.I2CReadEx(offset, data_count) # read content
        except:
            data: list[int] = []
        return data

    def _auto_set_bus_master(self) -> I2CMasterType:
        detected_master = I2CMasterType.AUTO_DETECT

        if self._instance_id == I2CInstances.CONTROLLER:
            if self.spm_ctrl.get_controller_type() == self.spm_ctrl.ControllerType.CX:
                if self.spm_ctrl.get_firmware_type() == self.spm_ctrl.FirmwareType.LINUX:
                    if self._bus_id == I2CBusID.User:
                        detected_master = I2CMasterType.EMBEDDED_AVALON
                    else:
                        detected_master = I2CMasterType.EMBEDDED_LINUX
                else:
                    detected_master = I2CMasterType.EMBEDDED_AVALON
        elif self._instance_id == I2CInstances.MAIN_APP:
            detected_master = I2CMasterType.ACCESSORY_MASTER

        return detected_master

class I2CChip():

    def __init__(self, bus_addr: int, offset_mode: I2COffsetMode, name: str = "", bus_master: I2CBusMaster = None, auto_lock: bool = True):
        """ Minimal initialization is bus_addr and offset_mode. connection to bus master can be done later by bus_master.assign_chip() """
        self._bus_master = bus_master
        self._chip_ref = -1
        self.name = name
        self.bus_address = bus_addr
        self.offset_mode = offset_mode
        self.auto_lock = auto_lock
        if self._bus_master is not None:
            self._bus_master.assign_chip(self)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name:str):
        self.__name = name

    @property
    def bus_address(self) -> int:
        return self.__bus_addr

    @bus_address.setter
    def bus_address(self, addr: int):
        self.__bus_addr = addr

    @property
    def offset_mode(self) -> I2COffsetMode:
        return self.__offset_mode

    @offset_mode.setter
    def offset_mode(self, mode: I2COffsetMode):
        self.__offset_mode = mode

    @property
    def auto_lock(self) -> bool:
        return self.__auto_lock

    @auto_lock.setter
    def auto_lock(self, lock:bool):
        self.__auto_lock = lock

    def setup_bus_connection(self, bus_master: I2CBusMaster, chip_ref: int):
        self._bus_master = bus_master
        self._chip_ref = chip_ref

    def activate(self):
        self._bus_master.setup_metadata(self.bus_address, self.offset_mode, self.auto_lock)

    def get_chip_ref(self) -> int:
        return self._chip_ref

    def get_bus(self):
        self._bus_master.activate_chip(self)

    def is_connected(self) -> bool:
        return self._bus_master.check_connection(self)

    def write_bytes_with_offset(self, offset: int, data: list[int]) -> bool:
        self.get_bus()
        done = self._bus_master.write_bytes(offset, data)
        return done

    def write_byte_with_offset(self, offset:int, data:int) -> bool:
        return self.write_bytes_with_offset(offset, [data])

    def write_bytes(self, data: list[int]) -> bool:
        return self.write_bytes_with_offset(0, data)

    def write_byte(self, data: int) -> bool:
        return self.write_bytes_with_offset(0, [data])

    def read_bytes_with_offset(self, offset:int, count: int) -> list[int]:
        self.get_bus()
        data = self._bus_master.read_bytes(offset, count)
        return data

    def read_byte_with_offset(self, offset: int) -> int:
        data = self.read_bytes_with_offset(offset, count=1)
        return data[0]

    def read_bytes(self, count: int) -> list[int]:
        data = self.read_bytes_with_offset(0, count)
        return data

    def read_byte(self) -> int:
        data = self.read_bytes_with_offset(0, count=1)
        return data[0]




""" The fit module where the obtained sweep data gets fitted with simple harmonic oscillator funciton
Copyright Nanosurf AG 2021
License - MIT
"""
import math
from pickle import FALSE
import numpy as np
from scipy import optimize as sci_optimize
from nanosurf.lib.datatypes import nsf_thread
from app import module_base


class FrequencySweepFitData():
    def __init__(self):
        self.result_ok = False
        self.result_fit_freq = []
        self.result_fit_amplitude = []
        self.result_fit_phase = []
        self.result_resonance_frequency_amplitude = 0.0
        self.result_q_factor_amplitude = 0.0

class FrequencySweepFitWorker(nsf_thread.NSFBackgroundWorker):

    par_measured_amplitudes = []
    par_measured_phases = []
    par_measured_frequencys = []
    par_init_param_amplitude = [0, 0, 0, 0, 0]
    par_init_param_phase = [0, 0, 0, 0]

    def __init__(self, my_module: module_base.ModuleBase):
        super().__init__(my_module)
        self.resulting_data = FrequencySweepFitData()
        
    def do_work(self):
        self.resulting_data = FrequencySweepFitData()
        self.resulting_data.result_ok = True
        
        ok = self.estimate_initial_parameter()
        self.resulting_data.result_ok &= ok
        
        ok, *_ = self.fit_amplitude()
        self.resulting_data.result_ok &= ok
        
        ok, *_ = self.fit_phase()
        self.resulting_data.result_ok &= ok

    def get_result(self) -> FrequencySweepFitData:
        return self.resulting_data

    def phase_function(self, x, fn, q, a, b) -> list[float]:
        """ Phase response of a damped driven harmonic oscillator """
        return np.arctan(q * (fn**2 - x**2) / (fn * x)) + a*x + b

    def amplitude_function(self, x, fn, q, a, b, c) -> list[float]:
        """ Amplitude response of a damped driven harmonic oscillator"""
        return a / np.sqrt((fn**2 - x**2) ** 2 + ((fn * x) / q)**2) + b*x + c

    def estimate_initial_parameter(self) -> bool:
        print("estimating initial parameter for curve fitting...")
        preparation_ok = False
        amplitudes = self.par_measured_amplitudes
        phases = self.par_measured_phases
        frequencies = self.par_measured_frequencys
        
        try:
            # Get max amplitude frequency from raw data
            fn_est = frequencies[np.argmax(amplitudes)]

            # Calculate q-factor from estimated resonance freq and bandwidth
            bw = 2.0*abs((fn_est - frequencies[round((np.argmax(amplitudes) - np.argmin(amplitudes))/2.0 + np.argmin(amplitudes))]))
            q = fn_est / bw

            # Calculate the linear fit to take the influence of other vibration into account
            slope_fit, offset_fit = np.polyfit(frequencies, amplitudes, 1)

            # Calculate the excitation amplitude from max amplitude and max amplitude frequency
            a_excitations = amplitudes*np.sqrt(np.square(fn_est**2 - np.square(frequencies)) + np.square(fn_est*frequencies/q)) + slope_fit*frequencies + offset_fit
            amplitude_excitation = np.mean(a_excitations)

            # Calculate the linear fit to get the slope
            phase_slope, *_ = np.polyfit(frequencies, phases, 1)
            phase_offset = np.mean(phases)

            self.par_init_param_amplitude[0] = fn_est
            self.par_init_param_amplitude[1] = q
            self.par_init_param_amplitude[2] = amplitude_excitation
            self.par_init_param_amplitude[3] = slope_fit
            self.par_init_param_amplitude[4] = offset_fit

            self.par_init_param_phase[0] = fn_est
            self.par_init_param_phase[1] = q
            self.par_init_param_phase[2] = phase_slope
            self.par_init_param_phase[3] = phase_offset
            # print(self.par_init_param_amplitude)
            # print(self.par_init_param_phase)

            # Get resonance frequency and Q-factor from phase fit as initial parameter for amplitude fit
            ok, p_opt = self.fit_phase()
            if ok:
                self.par_init_param_amplitude[0] = p_opt[0]
                self.par_init_param_amplitude[1] = p_opt[1]
                print("initial parameter estimating successful")
                preparation_ok = True
        except:
            self.logger.warn("initial parameter estimating not successful")
        return preparation_ok

    def fit_phase(self) -> tuple[bool, list[float]]:
        print("fitting phase data...")
        initial = self.par_init_param_phase
        frequencies = np.asarray(self.par_measured_frequencys)
        phases =  np.asarray(self.par_measured_phases)

        ok = False
        p_opt:list[float] = [] 
        try:
            phases_rad = phases*math.pi/180.0
            p_opt, *_ = sci_optimize.curve_fit(self.phase_function, frequencies, phases_rad, p0=initial)

            fn, q, phase_slope, phase_offset = p_opt
            phase_response = self.phase_function(frequencies, fn, q, phase_slope, phase_offset)

            phase_array_deg = np.asarray(phase_response) * 180.0 / math.pi
            self.resulting_data.result_fit_phase = phase_array_deg

            ok = True
            print("phase fit successful")
        except:
            self.logger.warn("Could not fit phase")
            ok = False
        return (ok, p_opt)

    def fit_amplitude(self) -> tuple[bool, list[float]]:
        print("fitting amplitude data...")
        initial = self.par_init_param_amplitude
        freq = self.par_measured_frequencys
        amplitude = self.par_measured_amplitudes

        ok = False
        p_opt:list[float] = [] 
        try:
            p_opt, *_= sci_optimize.curve_fit(self.amplitude_function, freq, amplitude, p0=initial)
            fn, q, amplitude_excitation, amplitude_slope, amplitude_offset  = p_opt

            self.resulting_data.result_fit_freq = freq
            self.resulting_data.result_resonance_frequency_amplitude = fn
            self.resulting_data.result_q_factor_amplitude = q

            amplitude_at_peak_freq = self.amplitude_function(freq, fn, q, amplitude_excitation, amplitude_slope, amplitude_offset)
            self.resulting_data.result_fit_amplitude = amplitude_at_peak_freq 
            print("amplitude fit successful")
            ok = True
        except:
            self.logger.warn("Could not fit amplitude")
            ok = False
        return (ok, p_opt)


""" This is the screen of the module
Copyright Nanosurf AG 2021
License - MIT
"""
import numpy as np
from PySide2 import QtWidgets
import nanosurf as nsf
from app import module_base
from modules.scan_module import module, settings

class ResultTableID(nsf.gui.TableEntryIDs):
    """ identifier id are used in a nsf_table widget"""
    Items = 0
    Marker_X = 1
    Marker_Y = 2
    Noise_Floor = 3

ComboChannelEntries = [
    nsf.gui.NSFComboEntry(settings.ChannelD.Deflection,'Deflection'),
    nsf.gui.NSFComboEntry(settings.ChannelD.Topography,'Topography'),
]

""" some useful list of allowed prefixes used by nsf_sci_edit widgets"""
allowed_count_units = [nsf.sci_val.up.Prefix.base]
allowed_time_units = [nsf.sci_val.up.Prefix.base, nsf.sci_val.up.Prefix.milli]
allowed_meter_units = [nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro, nsf.sci_val.up.Prefix.nano]

class ScanScreen(module_base.ModuleScreen):
    def __init__(self):
        super().__init__()

    def do_setup_screen(self, module: module.ScanModule):
        """ create here your gui with all controls and their layout"""
        self.module = module

        # left layout - main controls ------------------------------------------------------------

        self.scival_image_size = nsf.gui.NSFSciEdit("Image Size")
        self.scival_image_size.set_allowed_prefix_ids(allowed_meter_units)
        self.scival_image_size.set_prefix_id(nsf.sci_val.up.Prefix.micro)
        self.scival_image_size.set_precision(2)
        self.scival_image_size.set_value_min_max(0.0, 1.0)

        self.scival_time_per_line = nsf.gui.NSFSciEdit("Time per Line")
        self.scival_time_per_line.set_allowed_prefix_ids(allowed_time_units)
        self.scival_time_per_line.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_time_per_line.set_precision(2)
        self.scival_time_per_line.set_value_min_max(0.01, 5.0)

        self.scival_points = nsf.gui.NSFSciEdit("Data points")
        self.scival_points.set_allowed_prefix_ids(allowed_count_units)
        self.scival_points.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_points.set_precision(0)
        self.scival_points.set_value_min_max(2, 8000)
        
        self.label_channel_dir = QtWidgets.QLabel()
        self.combo_channel_dir = nsf.gui.NSFComboBox(ComboChannelEntries,"Channel")
        
        self.button_start_stop = QtWidgets.QPushButton("Start")

        self.layout_left = QtWidgets.QVBoxLayout()
        self.layout_left.addWidget(self.scival_image_size)
        self.layout_left.addWidget(self.scival_time_per_line)
        self.layout_left.addWidget(self.scival_points)
        self.layout_left.addWidget(self.label_channel_dir)
        self.layout_left.addWidget(self.combo_channel_dir)
        self.layout_left.addStretch()
        self.layout_left.addWidget(self.button_start_stop)

        # mid layout - plots and result ---------------------------------------------------------
        self.chart_plot = nsf.gui.NSFChart()
        self.colormap_plot = nsf.gui.NSFColormap()
        self.spec_plot = nsf.gui.NSFChart(logmodex=True, logmodey=True)

        self.check_show_histogram = QtWidgets.QCheckBox("Show Histogram")
        self.check_show_backward = QtWidgets.QCheckBox("Show Backward line")
        self.check_show_power_spec = QtWidgets.QCheckBox("Show Spectral density")
        self.check_show_compressed_spec = QtWidgets.QCheckBox("Compress Spectrum")
        self.tableResults = nsf.gui.NSFNameValueTable(ResultTableID)
        self.tableResults.define_entry(ResultTableID.Items,"Current line")
        self.tableResults.define_entry(ResultTableID.Marker_X,"Peak X")
        self.tableResults.define_entry(ResultTableID.Marker_Y,"Peak Y")
        self.tableResults.define_entry(ResultTableID.Noise_Floor,"Noise floor")

        self.layout_mid  = QtWidgets.QVBoxLayout()
        self.layout_color_plot  = QtWidgets.QHBoxLayout()
        self.layout_color_plot.addWidget(self.colormap_plot)
        self.layout_color_plot.addWidget(self.colormap_plot.histogram)
        self.layout_colormap  = QtWidgets.QVBoxLayout()
        self.layout_colormap.addLayout(self.layout_color_plot)
        self.layout_colormap.addWidget(self.check_show_histogram)

        self.layout_line_chart  = QtWidgets.QVBoxLayout()
        self.layout_line_chart.addWidget(self.chart_plot)
        self.layout_line_chart.addWidget(self.check_show_backward)
        self.layout_spec_chart  = QtWidgets.QVBoxLayout()
        self.layout_spec_chart.addWidget(self.spec_plot)
        self.layout_spec_chart.addWidget(self.check_show_power_spec)
        self.layout_spec_chart.addWidget(self.check_show_compressed_spec)
        self.layout_charts  = QtWidgets.QVBoxLayout()
        self.layout_charts.addLayout(self.layout_line_chart)
        self.layout_charts.addLayout(self.layout_spec_chart)
        self.layout_plots  = QtWidgets.QHBoxLayout()
        self.layout_plots.addLayout(self.layout_colormap)
        self.layout_plots.addLayout(self.layout_charts)
        self.layout_mid.addLayout(self.layout_plots)
        self.layout_mid.addWidget(self.tableResults)

        # right layout - additional user inputs

        self.layout_right= QtWidgets.QVBoxLayout()
        # self.layout_right.addSpacerItem(app_gui.StdVSpacer())

        # set GUI controls
        self.screen_layout = QtWidgets.QHBoxLayout()
        # stretch only plot area and keep controls fix in size
        self.screen_layout.addLayout(self.layout_left, 0)
        self.screen_layout.addLayout(self.layout_mid,  1) 
        self.screen_layout.addLayout(self.layout_right,0)
        self.setLayout(self.screen_layout)

        self.bind_gui_elements()
        self.init_plot()
        self.enter_gui_state_idle()

    def bind_gui_elements(self):
        """ connect here all gui widgets to settings of the module or any other source"""
        
        # binding ProVal to widgets ensure that they are alway in sync 
        nsf.gui.connect_to_property(self.combo_channel_dir, self.module.settings.channel_id)
        nsf.gui.connect_to_property(self.scival_image_size, self.module.settings.image_size)
        nsf.gui.connect_to_property(self.scival_time_per_line, self.module.settings.time_per_line)
        nsf.gui.connect_to_property(self.scival_points, self.module.settings.points_per_line)
        nsf.gui.connect_to_property(self.check_show_backward, self.module.settings.show_backward)
        nsf.gui.connect_to_property(self.check_show_power_spec, self.module.settings.show_power_spec)
        nsf.gui.connect_to_property(self.check_show_compressed_spec, self.module.settings.show_compress_spec)
        
        self.check_show_histogram.stateChanged.connect(self._show_histogram_clicked)
        self.check_show_backward.stateChanged.connect(self._show_backward_clicked)
        self.check_show_power_spec.stateChanged.connect(self._show_power_spec_clicked)
        self.check_show_compressed_spec.stateChanged.connect(self._show_compressed_spec_clicked)

        # buttons have to be connected separately
        self.button_start_stop.clicked.connect(self.on_button_start_stop_clicked)    
        
        # listen to signals from the core module to react and update the gui
        self.module.sig_work_start_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_stop_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_active.connect(self.enter_gui_state_active)
        self.module.sig_work_done.connect(self.enter_gui_state_idle)
        self.module.sig_new_data_available.connect(self.show_new_data)
        self.module.sig_data_invalid.connect(self.set_data_invalid)

    def init_plot(self):
        channel_name = self.combo_channel_dir.entry_name(self.module.settings.channel_id.value)

        self.colormap_plot.set_title(f"Scan image - {channel_name}")
        self.colormap_plot.set_label(nsf.gui.NSFColormap.Axis.bottom, "X-Axis")
        self.colormap_plot.set_unit(nsf.gui.NSFColormap.Axis.bottom, "m")
        self.colormap_plot.set_label(nsf.gui.NSFColormap.Axis.left, "Y-Axis")
        self.colormap_plot.set_unit(nsf.gui.NSFColormap.Axis.left, "m")  
        self.colormap_plot.set_unit(nsf.gui.NSFColormap.Axis.z, "m")  

        scan_line = np.linspace(0, self.module.settings.image_size.value, self.module.settings.points_per_line.value)
        self.colormap_plot.set_xy_range(scan_line, scan_line)
        self.colormap_plot.set_data_points(self.module.settings.points_per_line.value, self.module.settings.points_per_line.value)

        self.chart_plot.set_title(f"Current line chart - {channel_name}")
        self.chart_plot.set_label(nsf.gui.NSFChart.Axis.bottom, "X-Axis")
        self.chart_plot.set_unit(nsf.gui.NSFChart.Axis.bottom, "m")
        self.chart_plot.set_label(nsf.gui.NSFChart.Axis.left, "Data")
        self.chart_plot.set_unit(nsf.gui.NSFChart.Axis.left, "m")  
        self.chart_plot.plot.setXRange(0, self.module.settings.image_size.value)
        self.chart_plot.clear_plots()

        self.spec_plot.set_title(f"Spectrum of - {channel_name}")
        self.spec_plot.set_label(nsf.gui.NSFChart.Axis.bottom, "Frq")
        self.spec_plot.set_unit(nsf.gui.NSFChart.Axis.bottom, "Hz")
        self.spec_plot.set_label(nsf.gui.NSFChart.Axis.left, "Amplitude")
        self.spec_plot.set_unit(nsf.gui.NSFChart.Axis.left, "")  
        self.spec_plot.clear_plots()

    def on_button_start_stop_clicked(self):
        if self.module.is_worker_busy():
            self.module.stop_worker()
        else:
            self.module.start_worker()

    def _show_histogram_clicked(self):
        if self.check_show_histogram.isChecked():
            self.colormap_plot.histogram.show()
        else:
            self.colormap_plot.histogram.hide()

    def _show_backward_clicked(self):
        self.update_plot()

    def _show_power_spec_clicked(self):
        self.update_plot()

    def _show_compressed_spec_clicked(self):
        self.update_plot()

    def enter_gui_state_wait(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=True)

    def enter_gui_state_active(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())
        self.init_plot()

    def enter_gui_state_idle(self):
        self.set_parameter_widget_enable_state(enabled=True)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())

    def set_parameter_widget_enable_state(self, enabled: bool = True):
        self.scival_image_size.setEnabled(enabled)
        self.scival_time_per_line.setEnabled(enabled)
        self.scival_points.setEnabled(enabled)
        self.combo_channel_dir.setEnabled(enabled)

    def start_stop_button_state(self, wait: bool = False, stop_state: bool = False):
        if wait:
            self.button_start_stop.setEnabled(False)
            self.button_start_stop.setText("Wait...")
        else:
            self.button_start_stop.setEnabled(True)
            self.button_start_stop.setText("Stop" if stop_state else "Start")

    def show_new_data(self):
        self.update_plot()
        self.update_result()

    def set_data_invalid(self):
        self.chart_plot.clear_plots()
        self.tableResults.clear_values()

    def update_result(self):
        res = self.module.get_worker_result()
        self.tableResults.set_value(ResultTableID.Items, res.scan_line_index, "", precision=0)

    def update_plot(self): 
        current_data = self.module.get_worker_result()
        if current_data.stream.get_stream_length() > 1:
            # color map update
            self.colormap_plot.plot_channel(current_data.stream.get_channel(0), current_data.scan_line_index)

            self.chart_plot.plot_stream(current_data.stream, 0, layer_index=0)
            
            if self.module.settings.show_backward.value:
                self.chart_plot.plot_stream(current_data.stream, 1, layer_index=1)
            else:
                self.chart_plot.plot_data([], layer_index=1)
            
            calc_power_spec = self.module.settings.show_power_spec.value
            compress_spec = self.module.settings.show_compress_spec.value
            samplefrq = current_data.stream.get_stream_length()/self.module.settings.time_per_line.value
            spec = nsf.sci_math.calc_fft(current_data.stream.get_channel(0), samplerate=samplefrq, powerspectrum=calc_power_spec, compress=compress_spec)
            found, peak_x ,peak_y = nsf.sci_math.find_highest_peak(spec)
            if found:
                self.spec_plot.set_marker(peak_x, peak_y)
                self.tableResults.set_value(ResultTableID.Marker_X, peak_x, spec.get_stream_unit())
                self.tableResults.set_value(ResultTableID.Marker_Y, peak_y, spec.get_channel(0).unit)
            else:
                self.spec_plot.clear_marker()
                self.tableResults.clear_value(ResultTableID.Marker_X)
                self.tableResults.clear_value(ResultTableID.Marker_Y)
            
            noise_floor = nsf.sci_math.get_noise_floor(spec)
            self.spec_plot.plot_stream(spec)
            self.spec_plot.plot_data(y=[noise_floor for i in range(spec.get_stream_length())], x=spec.get_stream_range().value, layer_index=1)
            self.tableResults.set_value(ResultTableID.Noise_Floor, noise_floor, spec.get_channel(0).unit)



"""scientific value library formats and parse nice looking number strings into float numbers
    It handles units and exponent characters e.g. A string "2.45223kg" result in a float number 2452.23 and unit 'g'
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import sys
import math
import re

import nanosurf.lib.datatypes.sci_val.unit_prefix as up

class Token:
    success = False,
    value = 0,
    unit = "",
    prefix_id = up.Prefix.base

def to_value(val_str: str, current_unit: str = "", current_prefix_id: up.Prefix = up.Prefix.auto_, use_current_unit: bool = False) -> Token:
    instance = SciValParser(val_str, current_unit, current_prefix_id, use_current_unit=use_current_unit)
    return instance.str_to_value()

def to_string(value: float, unit: str, precision : int = 3, prefix_id:up.Prefix = up.Prefix.auto_, allowed_prefix_ids: list[up.Prefix] = up.default_prefix_range):
    instance = SciValFormatter(value,precision, prefix_id,allowed_prefix_ids,unit)
    return instance.value_to_str()

class SciValFormatter:
    def __init__(self, value: float, precision: int, prefix_id: up.Prefix, allowed_prefix_ids: list[up.Prefix], unit_str: str):
        self._value = value
        self._precision = precision
        self._prefix_id = prefix_id
        self._allowed_prefix_ids = allowed_prefix_ids
        self._unit = unit_str
        self._config_formatter()
        
    def value_to_str(self):
        self._convert_value_to_string()
        self._append_unit()
        return self._result_string

    def _config_formatter(self):
        self._result_string = ""
        self._exponent_value = 0
        self._no_prefix = (self._unit == "%") or (self._unit == "") or (self._unit == "°")
        self._no_space = (self._unit == "%") or (self._unit == "°")
        self._is_negative = (self._value < 0.0)
        if self._no_prefix:
            self._prefix_id = up.Prefix.base

    def _convert_value_to_string(self):
        self._config_formatter()
        if self._is_value_on_limit():
            self._result_string = self._create_inf_string()
        else:
            self._format_with_prefix()
            self._limit_lower_and_upper_bounds()

    def _format_with_prefix(self):
        if self._prefix_id == up.Prefix.auto_:
            self._format_with_auto_prefix()
        else:
            self._format_with_fixed_prefix()

    def _format_with_auto_prefix(self):
        self._exponent_value  = self._calculate_exponent(self._value)   
        auto_prefix_id =  up.prefix_value_to_prefix_id(self._exponent_value)     
        exponent_absolute_value = up.prefix_id_to_absolute_value(auto_prefix_id)
        self._value /= exponent_absolute_value
        self._result_string = self._format(self._value, self._precision)
        bigger_than_epsilon = abs(self._value) > sys.float_info.epsilon
        if (self._value > -1.0) and (self._value < +1.0) and bigger_than_epsilon and (self._exponent_value >= (up.prefix_min_exp_value+3)):
            self._decrement_exponent()    
        self._switch_to_closest_allowed_prefix(up.prefix_value_to_prefix_id(self._exponent_value))            

    def _format_with_fixed_prefix(self):
        exponent_absolute_value = up.prefix_id_to_absolute_value(self._prefix_id)
        self._exponent_value  = up.prefix_id_to_exponent_value(self._prefix_id)
        self._value /= exponent_absolute_value
        self._result_string = self._format(self._value, self._precision)

    def _format(self, value, precision):
        format_str = "{val:."+str(precision)+"f}"
        return format_str.format(val= value)

    def _calculate_exponent(self, value):
        tmp_exponent = 1
        if value != 0.0:
            tmp_exponent = math.floor(math.log10(abs(value)))
            if (value > -1.0) and (value < +1.0):
               tmp_exponent += 1
        return int(int(tmp_exponent / 3) * 3)      

    def _decrement_exponent(self):
        self._value *= 1000.0
        self._exponent_value -= 3
        self._result_string = self._format(self._value, self._precision)

    def _increment_exponent(self):
        self._value /= 1000.0
        self._exponent_value += 3
        self._result_string = self._format(self._value, self._precision)

    def _switch_to_closest_allowed_prefix(self, start_prefix_id: up.Prefix):
        if start_prefix_id not in self._allowed_prefix_ids:
            diff = up.prefix_max_exp_value - up.prefix_min_exp_value
            new_prefix = up.Prefix.base
            for allowed_prefix_id in self._allowed_prefix_ids:
                if (abs(self._exponent_value - up.prefix_id_to_exponent_value(allowed_prefix_id))) < diff:
                    new_prefix_id = allowed_prefix_id
                    diff = abs(self._exponent_value - up.prefix_id_to_exponent_value(allowed_prefix_id))
            self._exponent_value = up.prefix_id_to_exponent_value(new_prefix_id)
            prev_prefix_factor = up.prefix_id_to_absolute_value(start_prefix_id)
            self._value = self._value * prev_prefix_factor / up.prefix_id_to_absolute_value(new_prefix_id)
            self._result_string = self._format(self._value, self._precision)

    def _limit_lower_and_upper_bounds(self):
        pass

    def _append_unit(self):
        if not self._no_space and not self._no_prefix:
            self._result_string += " "
        self._result_string += up.prefix_value_to_string(self._exponent_value)
        self._result_string += self._unit

    def _is_value_on_limit(self):
        limit = False
        limit = limit or math.isnan(self._value)
        limit = limit or math.isinf(self._value)
        return limit

    def _create_inf_string(self):
        if self._is_negative:
            return str(-math.inf)
        else:
            return str(math.inf)

class SciValParser:
    def __init__(self, val_str: str, current_unit: str, current_prefix_id: up.Prefix, use_current_unit: bool = True):
        self._str_value = val_str
        self._current_prefix_id = current_prefix_id
        self._current_unit = current_unit
        self._use_current_unit = use_current_unit
        self._string_regex = re.compile(r'(?:[\s]*)(?P<num>[-+]?[0-9]*\.?[0-9]*)(?:[\s]*)(?P<prefix>[\w]?)(?P<unit>[\w]*)')
        
    def str_to_value(self) -> Token:
        res = Token()
        res.success = False

        m = self._string_regex.match(self._str_value)
        if m:
            parsed_dict = m.groupdict()
            self._parsed_prefix =  parsed_dict['prefix']
            self._parsed_unit =  parsed_dict['unit']
     
            if self._need_parsed_prefix_unit_swap():
                t = self._parsed_prefix
                self._parsed_prefix = self._parsed_unit
                self._parsed_unit = t
            try:
                res.value = float(parsed_dict['num'])
                res.unit = self._parsed_unit
            except:
                res.value = 0.0
                res.unit = ""
            res.success = True

        if res.success:
            if self._use_current_unit == False:
                if not up.is_prefix_str_valid(self._parsed_prefix):
                    res.unit = self._parsed_prefix + self._parsed_unit     
                    prefix_id = up.Prefix.base
                else:
                    prefix_id = up.string_prefix_to_prefix_id(self._parsed_prefix)  
            elif self._use_no_prefix():
                prefix_id = up.Prefix.base
            elif self._used_parsed_prefix():
                prefix_id =  up.string_prefix_to_prefix_id(self._parsed_prefix)   
            elif self._use_current_prefix():
                prefix_id = self._current_prefix_id
            else:
                prefix_id = self._current_prefix_id
            res.prefix_id = prefix_id
            res.value *= up.prefix_id_to_absolute_value(res.prefix_id)
        return res

    def _need_parsed_prefix_unit_swap(self):
        if (self._parsed_prefix!="") and (self._parsed_unit==""): 
            return True
        return False

    def _use_no_prefix(self):
        if (self._current_unit == "") or (self._current_unit == "%") or (self._current_unit == "°") or (self._current_unit == "°C"):
            return True
        if (self._parsed_prefix!="") and (self._parsed_unit == "") and (self._parsed_prefix == self._current_unit) and (self._current_unit != up.prefix_id_to_string(self._current_prefix_id)):
            return True
        if (self._parsed_prefix + self._parsed_unit) == "px":
            return True
        return False

    def _used_parsed_prefix(self):
        if (self._parsed_unit == self._current_unit):
            return True
        parsed_combined = self._parsed_prefix + self._parsed_unit
        if (self._parsed_prefix!="") and (self._parsed_prefix != self._current_unit) and (parsed_combined != self._current_unit):
            return True
        return False

    def _use_current_prefix(self):
        if (self._parsed_prefix!="") and (self._parsed_unit == "") and (self._parsed_prefix == self._current_unit) and  (self._current_unit != up.prefix_id_to_string(self._current_prefix_id)):
            return True
        parsed_combined = self._parsed_prefix + self._parsed_unit
        if (self._parsed_prefix!="") and (self._parsed_unit != "") and (self._parsed_unit != self._current_unit) and (parsed_combined == self._current_unit):
            return True
        return False



"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import nanosurf.lib.devices.i2c.bus_master as i2c 

class Chip_TMP42X(i2c.I2CChip): 
    """ temperature sensor chip"""
    class Register(enum.IntEnum):
        LocalTempHiByte = 0x00
        RemoteTempHiByte1 = 0x01
        RemoteTempHiByte2 = 0x02
        RemoteTempHiByte3 = 0x03
        StatusReg = 0x08
        ConfigReg1 = 0x09
        ConfigReg2 = 0x0a
        ConversionRateReg = 0x0b
        OneShotStart = 0x0f
        LocalTempLowByte = 0x10
        RemoteTempLowByte1 = 0x11
        RemoteTempLowByte2 = 0x12
        RemoteTempLowByte3 = 0x13
        NCorrection1 = 0x21
        NCorrection2 = 0x22
        NCorrection3 = 0x23
        SoftReset = 0xfc
        ManufacturingID = 0xfe
        DeviceID = 0xff

    class ConversionRate(enum.IntEnum):
        Convert_every_16s = 0
        Convert_every_8s = 1
        Convert_every_4s = 2
        Convert_every_2s = 3
        Convert_1_per_sec = 4
        Convert_2_per_sec = 5
        Convert_4_per_sec = 6
        Convert_8_per_sec = 7

    class ConfigReg1(enum.IntEnum):
        ShutDown = 0x40
        TempRangeExtended = 0x04

    class ConfigReg2(enum.IntEnum):
        EnableSensor3 = 0x40
        EnableSensor2 = 0x20
        EnableSensor1 = 0x10
        EnableLocal   = 0x08
        EnableResistorCorrection = 0x04

    class StatusReg(enum.IntEnum):
        Busy = 0x80

    def __init__(self, bus_addr: int, **kwargs):
        super().__init__(bus_addr, offset_mode=i2c.I2COffsetMode.NoOffset,**kwargs)

    def start_measuring(self, rate: ConversionRate = ConversionRate.Convert_1_per_sec, active_sensors: ConfigReg2 = ConfigReg2.EnableLocal):
        self.ConversionRate = rate
        self.reg_config_reg2 = active_sensors
        self.reg_config_reg1 = 0x00

    @property
    def reg_config_reg1(self) -> int:
        self.write_byte(self.Register.ConfigReg1)
        return self.read_byte()

    @reg_config_reg1.setter
    def reg_config_reg1(self, val: ConfigReg1):
        self.write_bytes([self.Register.ConfigReg1, val])

    @property
    def reg_config_reg2(self) -> int:
        self.write_byte(self.Register.ConfigReg2)
        return self.read_byte()

    @reg_config_reg2.setter
    def reg_config_reg2(self, val: ConfigReg2):
        self.write_bytes([self.Register.ConfigReg2, val])

    @property
    def reg_conversion(self) -> int:
        self.write_byte(self.Register.ConversionRateReg)
        return self.read_byte()

    @reg_conversion.setter
    def reg_conversion(self, rate: ConversionRate):
        self.write_bytes([self.Register.ConversionRateReg, rate])

    @property
    def reg_local_temp_high(self) -> int:
        self.write_byte(self.Register.LocalTempHiByte)
        return self.read_byte()

    @property
    def reg_local_temp_low(self) -> int:
        self.write_byte(self.Register.LocalTempLowByte)
        return self.read_byte()
        
    @property
    def reg_status(self) -> int:
        self.write_byte(self.Register.StatusReg)
        return self.read_byte()

    @property
    def temperature(self) -> float:
        """ Get current chip temperature in degree. Read only"""
        temp =  float(self.reg_local_temp_high) + 0.0625 * int((self.reg_local_temp_low >> 4) & 0x0f)
        return temp 

    @property
    def is_busy(self) -> bool:
        return (self.reg_status & self.StatusReg.Busy) > 0

    

   


# studio_scripting_demo client.py

import nanosurf as nsf

studio = nsf.Studio()
if studio.connect():
    print(f"Available sessions: {studio.main.session.list()}")
    print(f"Connected with session = '{studio.session_id}'")

    # set a value 
    studio.spm.workflow.imaging.property.points_per_line.value = 200

    # set values with more direct access
    imaging = studio.spm.workflow.imaging
    imaging.property.scan_range_fast_axis.value = 2e-6
    imaging.property.scan_range_slow_axis.value = 2e-6

    # or even direct property shortening
    scan_mode = imaging.property.scan_mode
    scan_mode.value = scan_mode.EnumType.Single_Frame

    # call a workflow action
    studio.spm.workflow.imaging.start_imaging()

    # different methods to iterate over enum properties
    print("possible scan modes by list: "+str(imaging.property.scan_mode.enum))
    print("possible scan modes by enum: ")
    for mode in imaging.property.scan_mode.EnumType:
        print(mode)

    # converting a property to SciVal
    scan_range_fast_axis = imaging.property.scan_range_fast_axis
    scan_range = nsf.sci_val.SciVal(imaging.property.scan_range_fast_axis)
    print(f"Fast scan range converted to SciVal is {scan_range}")
    print(f"This give the same result: {imaging.property.scan_range_fast_axis.value}")
    print(f"And also this give the same result: {nsf.sci_val.convert.to_string(scan_range_fast_axis.value,scan_range_fast_axis.unit)}")
    print(f"And this give the same value but not so nicely formatted: {scan_range_fast_axis.value}{scan_range_fast_axis.unit}")

    studio.disconnect()
else:
    print(f"Connecting to studio failed: {studio.last_error}")




"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits

class CtrlUnitsSineWaveGenerator(ctrlunits._CtrlUnit):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm)
        self._lu_inst_name = lu_inst_name
        self._lu = self._spm.lowlevel.SignalAnalyzer(spm.lowlevel.SignalAnalyzer.Instance[lu_inst_name])
        self._lu.operating_mode.value = self._lu.OperatingMode.LockIn
        self._lu.target_amplitude_ctrl_mode.value = self._lu.AmplitudeCtrlMode.ConstDrive
 
    @property
    def inst_name(self):
        return self._lu_inst_name

    @property
    def amp(self):
        return self.amp_norm * self.amp_max

    @amp.setter
    def amp(self, val):
        self.amp_norm = val / self.amp_max

    @property
    def amp_norm(self):
        return self._lu.reference_amplitude.value / self._lu.reference_amplitude.value_max

    @amp_norm.setter
    def amp_norm(self, val):
        self._lu.reference_amplitude.value = val * self._lu.reference_amplitude.value_max
        self._lu.switch_to_target_amplitude_ctrl_mode()

    @property
    def amp_v(self):
        return self.amp_norm * self.target._voltage_range

    @amp_v.setter
    def amp_v(self, val):
        self.amp_norm = val / self.target._voltage_range

    @property
    def amp_max(self):
        return self.target.max

    @property
    def amp_unit(self):
        return self.target.unit

    @property
    def freq(self):
        return  self._lu.reference_frequency.value

    @freq.setter
    def freq(self, val):
        self._lu.reference_frequency.value = val

class CtrlUnitLockIn(CtrlUnitsSineWaveGenerator):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm, lu_inst_name)
        self.list_of_bandwidth = self._lu.DemodulatorBW
        self._lu.input.value = self._lu.Input.FastInDeflection
        self._fast_input_channel_index = {
            "DEFLECTION":spm.lowlevel.SignalAnalyzer.Input.FastInDeflection,
            "CH2":spm.lowlevel.SignalAnalyzer.Input.FastIn2,
            "USER":spm.lowlevel.SignalAnalyzer.Input.FastInUser,
        }
        try:
            # only v3.10.3 or newer has longer list of input channels
            self._hires_input_channel_index = {
                "DEFLECTION":spm.lowlevel.SignalAnalyzer.Input.InDeflection,
                "USER1":spm.lowlevel.SignalAnalyzer.Input.InUser1,
                "USER2":spm.lowlevel.SignalAnalyzer.Input.InUser2,
                "USER3":spm.lowlevel.SignalAnalyzer.Input.InUser3,
                "USER4":spm.lowlevel.SignalAnalyzer.Input.InUser4,
                "TIPCURRENT":spm.lowlevel.SignalAnalyzer.Input.InTipCurrent,
                "POSITIONX":spm.lowlevel.SignalAnalyzer.Input.InPositionX,
                "POSITIONY":spm.lowlevel.SignalAnalyzer.Input.InPositionY,
                "POSITIONZ":spm.lowlevel.SignalAnalyzer.Input.InPositionZ,
                "LATERAL":spm.lowlevel.SignalAnalyzer.Input.InLateral,
                "DETECTORSUM":spm.lowlevel.SignalAnalyzer.Input.InDetectorSum,
                "IN6":spm.lowlevel.SignalAnalyzer.Input.In6,
            }
        except:
            self._hires_input_channel_index = {
                "USER1":spm.lowlevel.SignalAnalyzer.Input.InUser1,
                "USER2":spm.lowlevel.SignalAnalyzer.Input.InUser2,
                "USER3":spm.lowlevel.SignalAnalyzer.Input.InUser3,
                "USER4":spm.lowlevel.SignalAnalyzer.Input.InUser4,
                "TIPCURRENT":spm.lowlevel.SignalAnalyzer.Input.InTipCurrent,
                "POSITIONX":spm.lowlevel.SignalAnalyzer.Input.InPositionX,
                "POSITIONY":spm.lowlevel.SignalAnalyzer.Input.InPositionY,
                "POSITIONZ":spm.lowlevel.SignalAnalyzer.Input.InPositionZ,
                "LATERAL":spm.lowlevel.SignalAnalyzer.Input.InLateral,
            }

    @property
    def reference_phase(self):
        return self._lu.reference_pase.value

    @reference_phase.setter
    def reference_phase(self, val):
        self._lu.reference_pase.value = val

    @property
    def demodulation_bw(self):
        return self._lu.demodulator_bw.value

    @demodulation_bw.setter
    def demodulation_bw(self, val):
        self._lu.demodulator_bw.value = val

    @property
    def input_amp(self):
        return self._lu.current_amplitude.value

    @property
    def input_amp_unit(self):
        return self._lu.current_amplitude.unit

    @property
    def input_amp_max(self):
        return self._lu.current_amplitude.value_max

    @property
    def input_phase(self):
        return self._lu.current_phase.value

    @property
    def input_phase_unit(self):
        return self._lu.current_phase.unit

    @property
    def input_phase_max(self):
        return self._lu.current_phase.value_max

    @property
    def source(self):
        return self.__source

    @source.setter
    def source(self, val):
        import nanosurf.lib.spm.lowlevel.ctrlunits.adc as adc
        import nanosurf.lib.spm.lowlevel.ctrlunits.channelmux as channelmux

        self.__source = val
        if isinstance(self.__source, adc.CtrlUnitADCFast):
            self.__source.target = self
            self._lu.input.value = self._fast_input_channel_index[self.__source.inst_name]
        elif isinstance(self.__source, adc.CtrlUnitADCHiRes):
            self.__source.target = self
            self._lu.input.value = self._hires_input_channel_index[self.__source.inst_name]
        elif isinstance(self.__source, channelmux.CtrlUnitLockInMux):
            self.__source.target = self
            if self.__source.channel != "":
                self.__source.channel =  self.__source.channel  # reset channel to saved value
        else:
            assert False, "Assigned Lock-In source not supported: type: "+str(type(val))
            self.__source = None
